# Copyright 2020 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import socket
import logging
import datetime
import traceback
from logging.handlers import DatagramHandler

# try to import sentry SDK
try:
    import sentry_sdk
except ImportError:
    sentry_sdk = None
try:
    from sentry_sdk.integrations.flask import FlaskIntegration
except ImportError:
    FlaskIntegration = None

from . import consts


class StructLogger(logging.Logger):
    initial_context = {}
    context = {}
    secrets_multi = []
    secrets_single = []

    def _log(self, level, msg, args, exc_info=None, extra=None, stack_info=False, **kwargs):  # pylint: disable=arguments-differ
        if not extra:
            extra = {}
        extra.update(StructLogger.context)
        extra.update(kwargs)
        extra = {k: v for k, v in extra.items() if v is not None}
        super()._log(level, msg, args, exc_info, extra, stack_info)

    def set_initial_ctx(self, **kwargs):
        if not StructLogger.initial_context:
            StructLogger.initial_context.update(kwargs)
            self.reset_ctx()

    def set_ctx(self, **kwargs):
        StructLogger.context.update(kwargs)

    def reset_ctx(self):
        StructLogger.context = StructLogger.initial_context.copy()

    def set_secrets(self, secrets):
        secrets_single = []
        secrets_multi = []
        for s in secrets:
            s = s.splitlines()
            if len(s) == 1:
                secrets_single.append(s[0])
            else:
                secrets_multi.append(s)
        secrets_single.sort(key=len, reverse=True)
        secrets_multi.sort(key=len, reverse=True)
        StructLogger.secrets_single = secrets_single
        StructLogger.secrets_multi = secrets_multi


logging.setLoggerClass(StructLogger)
root_logger = StructLogger('root', logging.WARNING)  # pylint: disable=invalid-name
logging.root = root_logger
logging.Logger.root = root_logger
logging.Logger.manager.root = root_logger


log = logging.getLogger(__name__)

g_clickhouse_handler = None
g_basic_logger_done = False
g_masking_handler = None


class MaskingLogRecord(logging.LogRecord):
    def __init__(self, *args, **kwargs):
        logging.LogRecord.__init__(self, *args, **kwargs)
        self._msg = None
        self._secrets = []

    def add_mask_secret(self, secret, where):
        self._secrets.append((secret, where))
        self._msg = None

    def _mask_secrets(self, msg):
        if not self._secrets:
            return msg

        for s, where in self._secrets:
            if where == 'start':
                msg2 = '******' + msg[len(s):]
            elif where == 'end':
                msg2 = msg[:-len(s)] + '******'
            else:  # middle
                msg2 = msg.replace(s, '******')

            msg = msg2

        return msg

    def getMessage(self):
        if self._msg:
            return self._msg
        msg = logging.LogRecord.getMessage(self)
        msg = self._mask_secrets(msg)
        self._msg = msg
        return self._msg


class MaskingHandler(logging.Handler):
    def __init__(self):
        super().__init__()
        self._buffer = []
        self._hanlders = {}

    def add_handler(self, name, handler):
        self._hanlders[name] = handler
        handler.setLevel(self.level)

    def remove_handler(self, name):
        if name in self._hanlders:
            del self._hanlders[name]

    def setLevel(self, level):
        super().setLevel(level)
        for handler in self._hanlders.values():
            handler.setLevel(level)

    def setFormatter(self, fmt):
        super().setFormatter(fmt)
        for handler in self._hanlders.values():
            handler.setFormatter(fmt)

    def flush(self):
        super().flush()
        for handler in self._hanlders.values():
            handler.flush()

    def close(self):
        super().close()
        for handler in self._hanlders.values():
            handler.close()

    def true_emit(self, record):
        for handler in self._hanlders.values():
            handler.emit(record)

    def emit(self, record):
        # if there are no secrets them emit the record now
        if not StructLogger.secrets_single and not StructLogger.secrets_multi:
            # if there are some buffered records then flush them now
            if self._buffer:
                for rec in self._buffer:
                    self.true_emit(rec)
                self._buffer = []
            self.true_emit(record)
            return

        # if there are multi line secrets then buffer records
        # and process them when the number of buffered records
        # is equal or greater then min number of lines of all secrets
        if StructLogger.secrets_multi:
            max_secrets_lines = len(StructLogger.secrets_multi[0])
            min_secrets_lines = len(StructLogger.secrets_multi[-1])

            self._buffer.append(record)

            if len(self._buffer) < min_secrets_lines:
                return

            matched = False
            s = None
            for s in StructLogger.secrets_multi:
                if len(s) > len(self._buffer):
                    continue

                # moving window over buffer if it is longer than secret
                for w in range(0, len(self._buffer) - len(s) + 1):

                    # match lines in secret and buffer window
                    for idx, line in enumerate(s):
                        rec = self._buffer[w + idx]
                        msg = rec.getMessage()

                        if idx == 0:
                            if not msg.endswith(line):
                                break
                        elif idx == len(s) - 1:
                            if not msg.startswith(line):
                                break
                            # matched whole secret, so perform masking
                            matched = True
                        elif msg != line:
                            #continue
                            break

                    if matched:
                        break

                if matched:
                    break

            if s and matched:
                # flush any lines before window
                for _ in range(0, w):
                    rec = self._buffer[0]
                    del self._buffer[0]
                    for ss in StructLogger.secrets_single:
                        rec.add_mask_secret(ss, 'middle')
                    self.true_emit(rec)

                # replace secret in affected lines
                for idx, line in enumerate(s):
                    rec = self._buffer[0]
                    del self._buffer[0]

                    if idx == 0:
                        rec.add_mask_secret(line, 'end')
                    elif idx == len(s) - 1:
                        rec.add_mask_secret(line, 'start')

                    # drop lines between first and last lines
                    if idx in [0, len(s) - 1]:
                        for ss in StructLogger.secrets_single:
                            rec.add_mask_secret(ss, 'middle')
                        self.true_emit(rec)

            elif len(self._buffer) == max_secrets_lines:
                # if number of buffered lines is equal to the number of the longest secret
                # then make a space for new line in a buffer
                rec = self._buffer[0]
                del self._buffer[0]
                for ss in StructLogger.secrets_single:
                    rec.add_mask_secret(ss, 'middle')
                self.true_emit(rec)

        # if there are only single line secrets then process them now
        if StructLogger.secrets_single and not StructLogger.secrets_multi:
            for ss in StructLogger.secrets_single:
                record.add_mask_secret(ss, 'middle')
            self.true_emit(record)

    def flush_log_entries(self):
        for rec in self._buffer:
            for ss in StructLogger.secrets_single:
                rec.add_mask_secret(ss, 'middle')
            self.true_emit(rec)

        self._buffer = []


class ClickhouseFormatter(logging.Formatter):
    def __init__(self, tags=None, fqdn=False):
        super().__init__()
        self.tags = tags if tags is not None else []

        if fqdn:
            self.host = socket.getfqdn()
        else:
            self.host = socket.gethostname()

    def get_extra_fields(self, record):
        # The list contains all the attributes listed in
        # http://docs.python.org/library/logging.html#logrecord-attributes
        skip_list = (
            'args', 'asctime', 'created', 'exc_info', 'exc_text', 'filename',
            'funcName', 'id', 'levelname', 'levelno', 'lineno', 'module',
            'msecs', 'msecs', 'message', 'msg', 'name', 'pathname', 'process',
            'processName', 'relativeCreated', 'thread', 'threadName', 'extra',
            # new to skip, not needed
            '_msg', '_secrets', 'stack_info')

        easy_types = (str, bool, dict, float, int, list, type(None))

        fields = {}

        for key, value in record.__dict__.items():
            if key not in skip_list:
                if isinstance(value, easy_types):
                    fields[key] = value
                else:
                    fields[key] = repr(value)

        if hasattr(record, 'extra'):
            for key, value in record.extra:
                if isinstance(value, easy_types):
                    fields[key] = value
                else:
                    fields[key] = repr(value)

        return fields

    def get_debug_fields(self, record):
        fields = {
            'stack_trace': self.format_exception(record.exc_info),
            'lineno': record.lineno,
            'process': record.process,
            'thread_name': record.threadName,
        }

        fields['funcName'] = record.funcName
        fields['processName'] = record.processName

        return fields

    @classmethod
    def format_source(cls, message_type, host, path):
        return "%s://%s/%s" % (message_type, host, path)

    @classmethod
    def format_timestamp(cls, time):
        tstamp = datetime.datetime.utcfromtimestamp(time)
        return tstamp.isoformat() + "Z"

    @classmethod
    def format_exception(cls, exc_info):
        return ''.join(traceback.format_exception(*exc_info)) if exc_info else ''

    @classmethod
    def serialize(cls, message):
        return bytes(json.dumps(message), 'utf-8')

    def format(self, record):
        msg = record.getMessage()
        if record.exc_info:
            for l in self.format_exception(record.exc_info).splitlines():
                msg += '\n' + l.rstrip()

        # Create message dict
        message = {
            '@timestamp': self.format_timestamp(record.created),
            # '@version': '1',
            'message': msg,
            'host': self.host,
            'path': record.pathname,
            'lineno': record.lineno,
            # 'tags': self.tags,

            # Extra Fields
            'level': record.levelname,
            'logger_name': record.name,
        }

        # Add extra fields
        message.update(self.get_extra_fields(record))

        # If exception, add debug info
        # TODO: for now skipped
        # if record.exc_info:
        #     message.update(self.get_debug_fields(record))

        return self.serialize(message)


class ClickhouseHandler(DatagramHandler):
    """Python logging handler for Clickhouse. Sends events over UDP.
    :param host: The host of the Clickhouse server.
    :param port: The port of the Clickhouse server (default 9001).
    :param fqdn; Indicates whether to show fully qualified domain name or not (default False).
    :param tags: list of tags for a logger (default is None).
    """

    def __init__(self, host, port=9001, tags=None, fqdn=False):
        super().__init__(host, port)
        self.formatter = ClickhouseFormatter(tags, fqdn)

    def makePickle(self, record):
        return self.formatter.format(record)


def setup_logging(service, clickhouse_addr=None):
    global g_clickhouse_handler, g_basic_logger_done, g_masking_handler  # pylint: disable=global-statement

    logging.setLogRecordFactory(MaskingLogRecord)

    if not g_basic_logger_done:
        g_masking_handler = MaskingHandler()
        sh = logging.StreamHandler()
        g_masking_handler.add_handler('stream', sh)
        logging.basicConfig(format=consts.LOG_FMT, level=logging.INFO, handlers=[g_masking_handler])
        g_basic_logger_done = True

    l = logging.getLogger()

    if g_clickhouse_handler:
        g_masking_handler.remove_handler('clickhouse')
        g_clickhouse_handler = None

    ch_addr = os.environ.get('KRAKEN_CLICKHOUSE_ADDR', None)
    if not ch_addr:
        ch_addr = clickhouse_addr
    if not ch_addr:
        ch_addr = consts.DEFAULT_CLICKHOUSE_ADDR
    host, port = ch_addr.split(':')
    g_clickhouse_handler = ClickhouseHandler(host, int(port), fqdn=True)
    l.set_initial_ctx(service=service)
    g_masking_handler.add_handler('clickhouse', g_clickhouse_handler)
    log.info('setup logging on %s to clickhouse: %s', service, ch_addr)


def setup_sentry(sentry_url):
    if not sentry_url or not sentry_sdk:
        return

    if FlaskIntegration:
        sentry_sdk.init(
            dsn=sentry_url,
            integrations=[FlaskIntegration()])
    else:
        sentry_sdk.init(
            dsn=sentry_url)

    log.info('sentry setup, DSN: %s...%s', sentry_url[:15], sentry_url[-5:])



"""extended tool

Revision ID: af348b43d9d5
Revises: 5e38f55beb11
Create Date: 2022-05-14 19:48:36.262751

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'af348b43d9d5'
down_revision = '5e38f55beb11'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('tools', sa.Column('location', sa.UnicodeText(), nullable=True))
    op.add_column('tools', sa.Column('entry', sa.UnicodeText(), nullable=True))
    op.add_column('tools', sa.Column('version', sa.Integer(), nullable=True))


def downgrade():
    op.drop_column('tools', 'version')
    op.drop_column('tools', 'entry')
    op.drop_column('tools', 'location')



# Copyright 2020 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import hmac
import hashlib
import logging

from flask import Blueprint, request, abort
import dateutil.parser

from .models import Project
from .bg import jobs as bg_jobs
from . import kkrq

log = logging.getLogger(__name__)


### GITHUB #############

def handle_github_webhook(project_id):
    payload = request.get_data()
    log.info('GITHUB for project_id:%s, payload: %s', project_id, payload)

    # check event
    event = request.headers.get('X-GitHub-Event')
    if event is None:
        log.warning('missing github event type in request header')
        abort(400, "missing github event type in request header")
    log.info('EVENT %s', event)
    if event not in ['push', 'pull_request']:
        msg = 'unsupported event'
        log.info(msg)
        return msg, 204

    # check project
    project = Project.query.filter_by(id=project_id).one_or_none()
    if project is None:
        log.warning('cannot find project %s', project_id)
        abort(400, "Invalid project id")

    if not project.webhooks or not project.webhooks.get('github_enabled', False):
        log.info('webhooks from github disabled')
        abort(400, "webhooks from github disabled")

    # check secret
    my_secret = project.webhooks.get('github_secret', None)
    if my_secret:
        my_secret = bytes(my_secret, 'ascii')
    if my_secret is not None:
        github_sig = request.headers.get("X-Hub-Signature")
        if github_sig is None:
            log.warning('missing signature in request header')
            abort(400, "missing signature in request header")
        github_digest_parts = github_sig.split("=", 1)
        my_digest = hmac.new(my_secret, payload, hashlib.sha1).hexdigest()

        if len(github_digest_parts) < 2 or github_digest_parts[0] != "sha1" or not hmac.compare_digest(github_digest_parts[1], my_digest):
            log.warning('bad signature %s vs %s', github_sig, my_digest)
            abort(400, "Invalid signature")

    req = json.loads(payload)

    # trigger running the project flow via rq
    if event == 'push':
        trigger_data = dict(trigger='github-' + event,
                            ref=req['ref'],
                            before=req['before'],
                            after=req['after'],
                            repo=req['repository']['clone_url'],
                            pusher=req['pusher'],
                            commits=req['commits'])
    elif event == 'pull_request':
        if req['action'] not in ['opened', 'synchronize']:
            msg = 'unsupported action %s' % req['action']
            log.info(msg)
            return msg, 204

        if req['action'] == 'opened' and req['pull_request']['commits'] == 0:
            msg = 'pull request with no commits, dropped'
            log.info(msg)
            return msg, 204

        if 'before' in req:
            before = req['before']
        else:
            before = req['pull_request']['base']['sha']

        if 'after' in req:
            after = req['after']
        else:
            after = req['pull_request']['head']['sha']

        trigger_data = dict(trigger='github-' + event,
                            action=req['action'],
                            pull_request=req['pull_request'],
                            before=before,
                            after=after,
                            repo=req['repository']['clone_url'],
                            sender=req['sender'])
    kkrq.enq(bg_jobs.trigger_flow, project.id, trigger_data)
    return "", 204


### GITEA #############

def handle_gitea_webhook(project_id):
    payload = request.get_data()
    log.info('GITEA for project_id:%s, payload: %s', project_id, payload)
    event = request.headers.get('X-Gitea-Event')
    signature = request.headers.get("X-Hub-Signature")
    log.info('GITEA event:%s, signature: %s', event, signature)
    return _handle_gitea_webhook(project_id, payload, event, signature)


def _handle_gitea_webhook(project_id, payload, event, signature):
    # check event
    if event is None:
        log.warning('missing gitea event type in request header')
        abort(400, "missing gitea event type in request header")
    log.info('EVENT %s', event)
    if event not in ['push', 'pull_request']:
        msg = 'unsupported event'
        log.info(msg)
        return msg, 204

    # check project
    project = Project.query.filter_by(id=project_id).one_or_none()
    if project is None:
        log.warning('cannot find project %s', project_id)
        abort(400, "Invalid project id")

    if not project.webhooks or not project.webhooks.get('gitea_enabled', False):
        log.info('webhooks from gitea disabled')
        abort(400, "webhooks from gitea disabled")

    # check secret
    my_secret = project.webhooks.get('gitea_secret', None)
    if my_secret:
        my_secret = bytes(my_secret, 'ascii')
    if my_secret is None:
        msg = 'secret not configured for gitea webhook'
        log.warning(msg)
        return msg, 204
    if signature is None:
        log.warning('missing signature in request header')
        abort(400, "missing signature in request header")
    digest_parts = signature.split("=", 1)
    my_digest = hmac.new(my_secret, payload, hashlib.sha1).hexdigest()

    if len(digest_parts) < 2 or digest_parts[0] != "sha1" or not hmac.compare_digest(digest_parts[1], my_digest):
        log.warning('bad signature %s vs %s', signature, my_digest)
        abort(400, "Invalid signature")

    req = json.loads(payload)

    # trigger running the project flow via rq
    if event == 'push':
        trigger_data = dict(trigger='gitea-' + event,
                            ref=req['ref'],
                            before=req['before'],
                            after=req['after'],
                            repo=req['repository']['clone_url'],
                            pusher=req['pusher'],
                            commits=req['commits'])
    elif event == 'pull_request':
        if req['action'] not in ['opened', 'synchronize', 'synchronized']:
            msg = 'unsupported action %s' % req['action']
            log.info(msg)
            return msg, 204

        if 'before' in req and 'after' in req:
            before = req['before']
            after = req['after']
        else:
            before = req['pull_request']['base']['sha']
            after = req['pull_request']['head']['sha']

        if after == before:
            msg = 'pull request with no commits, dropped'
            log.info(msg)
            return msg, 204

        trigger_data = dict(trigger='gitea-' + event,
                            action=req['action'],
                            pull_request=req['pull_request'],
                            before=before,
                            after=after,
                            repo=req['repository']['clone_url'],
                            sender=req['sender'])
    kkrq.enq(bg_jobs.trigger_flow, project.id, trigger_data)

    return "", 204


### GITLAB #############

def handle_gitlab_webhook(project_id):
    payload = request.get_data()
    log.info('GITLAB for project_id:%s, payload: %s', project_id, payload)
    event = request.headers.get('X-Gitlab-Event')
    token = request.headers.get('X-Gitlab-Token')
    return _handle_gitlab_webhook(project_id, payload, event, token)


def _handle_gitlab_webhook(project_id, payload, event, token):
    # check event
    if event is None:
        log.warning('missing gitlab event type in request header')
        abort(400, "missing gitlab event type in request header")
    log.info('EVENT %s', event)
    if event not in ['Push Hook', 'Merge Request Hook']:
        msg = 'unsupported event'
        log.info(msg)
        return msg, 204

    # check project
    project = Project.query.filter_by(id=project_id).one_or_none()
    if project is None:
        log.warning('cannot find project %s', project_id)
        abort(400, "Invalid project id")

    if not project.webhooks or not project.webhooks.get('gitlab_enabled', False):
        log.info('webhooks from gitlab disabled')
        abort(400, "webhooks from gitlab disabled")

    # check secret
    my_secret = project.webhooks.get('gitlab_secret', None)
    if my_secret is None:
        log.warning('secret not configured for gitea webhook')
        return "", 204
    if token is None:
        log.warning('missing token in request header')
        abort(400, "missing token in request header")

    if token != my_secret:
        log.warning('bad token %s vs %s', token, my_secret)
        abort(400, "Invalid token")

    req = json.loads(payload)

    # trigger running the project flow via rq
    if event == 'Push Hook':
        trigger_data = dict(trigger='gitlab-' + event,
                            ref=req['ref'],
                            before=req['before'],
                            after=req['after'],
                            repo=req['repository']['git_http_url'],
                            pusher=dict(full_name=req['user_name'],
                                        username=req['user_username'],
                                        email=req['user_email']),
                            commits=req['commits'])
    elif event == 'Merge Request Hook':
        obj = req['object_attributes']
        action = obj['action']
        if action not in ['open', 'update']:
            msg = 'unsupported action %s' % action
            log.info(msg)
            return msg, 204

        if 'oldrev' in obj:
            before = obj['oldrev']
        else:
            before = ''

        after = obj['last_commit']['id']

        if after == before:
            msg = 'merge request with no commits, dropped'
            log.info(msg)
            return msg, 204

        # get base url
        base_url = req['project']['web_url']
        base_url = base_url.rsplit('/', 2)[0]

        trigger_data = dict(trigger='gitlab-' + event,
                            action=action,
                            pull_request=dict(head=dict(ref=obj['source_branch']),
                                              base=dict(ref=obj['target_branch'],
                                                        sha=before),
                                              user=dict(login=req['user']['username'],
                                                        html_url='%s/%s' % (base_url, req['user']['username'])),
                                              html_url=obj['url'],
                                              number=obj['id'],
                                              updated_at=dateutil.parser.parse(obj['updated_at']).isoformat(),
                                              title=obj['title']),
                            before=before,
                            after=after,
                            repo=req['project']['git_http_url'],
                            sender=req['user'])
    kkrq.enq(bg_jobs.trigger_flow, project.id, trigger_data)

    return "", 204


### BLUEPRINT #############

def create_blueprint():
    bp = Blueprint('webhooks', __name__)

    bp.add_url_rule('/<int:project_id>/github', view_func=handle_github_webhook, methods=['POST'])
    bp.add_url_rule('/<int:project_id>/gitea', view_func=handle_gitea_webhook, methods=['POST'])
    bp.add_url_rule('/<int:project_id>/gitlab', view_func=handle_gitlab_webhook, methods=['POST'])

    return bp



"""added section to Artifact

Revision ID: a7de4bdee425
Revises: bb0d4908ba49
Create Date: 2020-02-16 13:50:35.889016

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'a7de4bdee425'
down_revision = 'bb0d4908ba49'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('artifacts', sa.Column('section', sa.Integer(), nullable=True))


def downgrade():
    op.drop_column('artifacts', 'section')



# Copyright 2021 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from sqlalchemy.exc import IntegrityError
from psycopg2.errors import UniqueViolation  # pylint: disable=no-name-in-module

from .. import utils
from ..models import db
from ..models import Agent, AgentAssignment


log = logging.getLogger(__name__)


def _create_agent(params, ag):
    now = utils.utcnow()
    params['last_seen'] = now
    params['authorized'] = True
    params['disabled'] = False

    try:
        a = Agent(**params)
        db.session.commit()
        log.info('created new agent %s', a)
    except Exception:
        db.session.rollback()
        a = Agent.query.filter_by(address=params['address']).one_or_none()
        log.info('using existing agent %s', a)
        if a:
            a.created = now
            if a.deleted:
                log.info('undeleting agent %s', a)
                a.deleted = None
            for f, val in params.items():
                setattr(a, f, val)
            db.session.commit()
        else:
            log.info('agent %s duplicated but cannot find it', params['address'])
            raise

    try:
        AgentAssignment(agent=a, agents_group=ag)
        db.session.commit()
    except IntegrityError as e:
        db.session.rollback()
        if not isinstance(e.orig, UniqueViolation):
            raise

    return a



# Copyright 2020 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging

import minio
from minio.lifecycleconfig import LifecycleConfig, Expiration, Rule
from minio.commonconfig import ENABLED, Filter
from minio.error import S3Error
import urllib3
from urllib3.exceptions import MaxRetryError

from . import consts
from .models import get_setting


log = logging.getLogger(__name__)


def get_minio_addr():
    minio_addr = get_setting('general', 'minio_addr')
    if not minio_addr:
        minio_addr = os.environ.get('KRAKEN_MINIO_ADDR', consts.DEFAULT_MINIO_ADDR)
    root_user = os.environ['MINIO_ROOT_USER']
    root_password = os.environ['MINIO_ROOT_PASSWORD']
    return minio_addr, root_user, root_password


def get_minio():
    minio_addr, root_user, root_password = get_minio_addr()
    http_client = urllib3.PoolManager(
        timeout=5,
        maxsize=10,
        cert_reqs='CERT_REQUIRED',
        retries=urllib3.Retry(
            total=3,
            backoff_factor=0.2,
            status_forcelist=[500, 502, 503, 504]
        )
    )
    mc = minio.Minio(minio_addr, access_key=root_user, secret_key=root_password, secure=False, http_client=http_client)
    return mc


def check_connection():
    minio_addr, root_user, root_password = get_minio_addr()
    http_client = urllib3.PoolManager(
        timeout=1,
        maxsize=10,
        cert_reqs='CERT_REQUIRED',
        retries=urllib3.Retry(
            total=2,
            backoff_factor=0.2,
            status_forcelist=[500, 502, 503, 504]
        )
    )
    mc = minio.Minio(minio_addr, access_key=root_user, secret_key=root_password, secure=False, http_client=http_client)
    try:
        mc.list_buckets()
    except MaxRetryError as ex:
        log.error('minio connection error: %s', ex)
        return False
    except S3Error as ex:
        log.error('minio connection error: %s', ex)
        return False

    return True


def get_or_create_minio_bucket_for_artifacts(branch_id):
    bucket_name = '%08d' % branch_id

    mc = get_minio()
    found = mc.bucket_exists(bucket_name)
    if not found:
        mc.make_bucket(bucket_name)

    return bucket_name


def get_or_create_minio_bucket_for_cache(job, step):
    bucket_name = '%08d-cache' % job.run.flow.branch_id

    if 'key' in step:
        cache_keys = [step['key']]
    else:
        cache_keys = step['keys']
    folders = {}
    for key in cache_keys:
        fldr = '%d-%s/%s' % (job.run.stage_id,
                             job.name.replace(' ', '_'),
                             key)
        folders[key] = fldr

    mc = get_minio()
    found = mc.bucket_exists(bucket_name)
    if not found:
        # create bucket
        mc.make_bucket(bucket_name)

        # set retention policy - delete files after 10 days
        cfg = LifecycleConfig(
            [
                Rule(
                    ENABLED,
                    rule_filter=Filter(prefix=""),
                    rule_id="rule-10-days",
                    expiration=Expiration(days=10),
                ),
            ],
        )
        mc.set_bucket_lifecycle(bucket_name, cfg)

    return bucket_name, folders


def get_or_create_minio_bucket_for_git(repo_url, branch_id=None, tool_id=None):
    if branch_id:
        bucket_name = '%08d-git' % branch_id
    elif tool_id:
        bucket_name = 'tool-%08d-git' % tool_id
    else:
        raise Exception('Branch or Tool id should not be None')

    folder = repo_url.replace('/', '_').replace(':', '_').replace('.', '_')

    mc = get_minio()
    found = mc.bucket_exists(bucket_name)
    if not found:
        mc.make_bucket(bucket_name)

    return bucket_name, folder


def get_or_create_minio_bucket_for_tool(tool_id):
    bucket_name = 'tool-%d' % tool_id

    mc = get_minio()
    found = mc.bucket_exists(bucket_name)
    if not found:
        mc.make_bucket(bucket_name)

    return bucket_name, mc



"""added new repo fields to stage

Revision ID: 2dd5cf2d12dc
Revises: cb7654d57bca
Create Date: 2021-01-02 15:03:35.913920

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '2dd5cf2d12dc'
down_revision = 'cb7654d57bca'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('stages', sa.Column('repo_error', sa.UnicodeText(), nullable=True))
    op.add_column('stages', sa.Column('repo_refresh_interval', sa.UnicodeText(), nullable=True))
    op.add_column('stages', sa.Column('repo_refresh_job_id', sa.UnicodeText(), nullable=True))
    op.add_column('stages', sa.Column('repo_state', sa.Integer(), nullable=True))
    op.add_column('stages', sa.Column('repo_version', sa.UnicodeText(), nullable=True))


def downgrade():
    op.drop_column('stages', 'repo_version')
    op.drop_column('stages', 'repo_state')
    op.drop_column('stages', 'repo_refresh_job_id')
    op.drop_column('stages', 'repo_refresh_interval')
    op.drop_column('stages', 'repo_error')



"""removed unique constraint agent non-deleted address

Revision ID: e0d8421619c2
Revises: e404f52562c2
Create Date: 2021-06-28 08:04:42.333490

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = 'e0d8421619c2'
down_revision = 'e404f52562c2'
branch_labels = None
depends_on = None


def upgrade():
    with op.get_context().autocommit_block():
        try:
            op.drop_index('ix_agents_address_not_deleted', table_name='agents')
        except Exception:
            pass


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index('ix_agents_address_not_deleted', 'agents', ['address'], unique=True)
    # ### end Alembic commands ###



# Copyright 2022 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from sqlalchemy.sql.expression import desc
import jsonschema

from .models import Tool
from . import minioops


def check_tool_schema(schema):
    try:
        jsonschema.validate(instance={}, schema=schema)
    except jsonschema.exceptions.SchemaError:
        raise
    except Exception:
        pass

    if "properties" not in schema:
        raise jsonschema.exceptions.SchemaError("Parameters of tool does not have 'properties' field.")

    required_fields = schema.get("required", [])
    if not isinstance(required_fields, list):
        raise jsonschema.exceptions.SchemaError("'required' field in tool fields definitions should have list value, not %s." % type(required_fields))


def create_or_update_tool(meta):
    name = meta['name']
    description = meta['description']
    location = meta['location']
    entry = meta['entry']
    fields = meta['parameters']
    version = meta.get('version', None)

    # check tool schema if it is ok
    check_tool_schema(fields)

    tool = None
    if version:
        # find tool to overwrite
        q = Tool.query
        q = q.filter_by(name=name)
        q = q.filter_by(version=version)
        q = q.filter_by(deleted=None)
        tool = q.one_or_none()
    else:
        # fint the latest tool to estimate next version
        q = Tool.query
        q = q.filter_by(name=name)
        q = q.filter_by(deleted=None)
        q = q.order_by(desc(Tool.created))
        tools = q.all()
        prev_tool = None
        for t in tools:
            if t.version[-1].isdigit():
                prev_tool = t
                break

    if not version:
        if not prev_tool:
            version = '1'
        else:
            m = re.match(r'^(.*)(\d+)$', prev_tool.version)
            if not m:
                version = '1'
            else:
                ver_base = m.group(1)
                ver_num = int(m.group(2))
                ver_num += 1
                version = '%s%d' % (ver_base, ver_num)

    if tool:
        tool.description = description
        tool.version = version
        tool.location = location
        tool.entry = entry
        tool.fields = fields
    else:
        tool = Tool(name=name, description=description, version=version, location=location, entry=entry, fields=fields)

    return tool


def store_tool_in_minio(tool_zip_fp, tool):
    bucket, mc = minioops.get_or_create_minio_bucket_for_tool(tool.id)
    dest = 'tool.zip'

    mc.put_object(bucket, dest, tool_zip_fp, -1, part_size=10*1024*1024, content_type="application/zip")

    tool.location = 'minio:%s/%s' % (bucket, dest)

    return tool



# Copyright 2020-2021 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import time
import logging

from flask import request
from sqlalchemy import or_
from sqlalchemy.orm import joinedload
from sqlalchemy.orm.attributes import flag_modified
from sqlalchemy.exc import IntegrityError
from psycopg2.errors import UniqueViolation  # pylint: disable=no-name-in-module
import giturlparse

from .models import db, Job, Step, Agent, TestCase, TestCaseResult, Issue, Secret, Artifact, File
from .models import System
from . import consts
from .bg import jobs as bg_jobs
from . import minioops
from .. import version
from . import kkrq
from . import utils
from . import dbutils

log = logging.getLogger(__name__)

JOB = {
    "name": "build-tarball",
    "id": 123,
    "steps": [{
        "tool": "git",
        "checkout": "git@gitlab.isc.org:isc-projects/kea.git",
        "branch": "master"
    }, {
        "tool": "shell",
        "cmd": "cd kea && autoreconf -fi && ./configure && make dist"
    }, {
        "tool": "artifacts",
        "type": "file",
        "upload": "aaa-{ver}.tar.gz"
    }]
}


def _left_time(job):
    now = utils.utcnow()
    slip = now - job.assigned
    timeout = job.timeout
    timeout2 = timeout - slip.total_seconds()
    # reduce slightly timeout
    timeout3 = timeout2 * 0.9
    log.info('%s-%s-%s: now: %s, slip:%s, to1: %s, to2: %s, to3: %s',
             job.name, job.system.id, job.agents_group_id,
             now, slip, timeout, timeout2, timeout3)
    return int(timeout3)


def _handle_get_job(agent):
    if agent.job is None:
        return {'job': {}}

    # handle canceling situation
    if agent.job.state == consts.JOB_STATE_COMPLETED:
        job = agent.job
        agent.job = None
        db.session.commit()
        log.info("unassigned canceled job %s from %s", job, agent)
        return {'job': {}}

    job = agent.job.get_json()

    job['timeout'] = _left_time(agent.job)

    # prepare test list for execution
    tests = []
    for tcr in agent.job.results:
        tests.append(tcr.test_case.name)
    if tests:
        job['steps'][-1]['tests'] = tests

    # attach trigger data to job
    if agent.job.run.flow.trigger_data:
        job['trigger_data'] = agent.job.run.flow.trigger_data.data
    elif agent.job.run.repo_data_id:
        # pick any repo for now (TODO: it should be more sophisticated and handle all repos)
        url = agent.job.run.repo_data.data[0]['repo']
        commits = agent.job.run.repo_data.data[0]['commits']
        job['trigger_data'] = [dict(repo=url,
                                    after=commits[0]['id'])]

    # attach storage info to job
    job['branch_id'] = agent.job.run.flow.branch_id
    job['flow_kind'] = agent.job.run.flow.kind
    job['flow_id'] = agent.job.run.flow_id
    job['run_id'] = agent.job.run_id

    minio_bucket = minioops.get_or_create_minio_bucket_for_artifacts(agent.job.run.flow.branch_id)
    minio_addr, minio_access_key, minio_secret_key = minioops.get_minio_addr()

    # prepare steps
    project = agent.job.run.flow.branch.project
    for step in job['steps']:
        add_minio = False

        # insert secret from ssh-key
        if 'ssh-key' in step:
            value = step['ssh-key']
            secret = Secret.query.filter_by(project=project, name=value, deleted=None).one_or_none()
            if secret is None:
                raise Exception("Secret '%s' does not exist in project %s" % (value, project.id))
            step['ssh-key'] = dict(username=secret.data['username'],
                                   key=secret.data['key'])

        # insert secret from access-token
        if 'access-token' in step:
            value = step['access-token']
            secret = Secret.query.filter_by(project=project, name=value).one_or_none()
            if secret is None:
                raise Exception("Secret '%s' does not exist in project %s" % (value, project.id))
            step['access-token'] = secret.data['secret']

        # custom fields for GIT
        if step['tool'] == 'git':
            # add http url to git
            url = step['checkout']
            url = giturlparse.parse(url)
            if url.valid:
                url = url.url2https
                step['http_url'] = url
            else:
                log.info('invalid git url %s', step['checkout'])

            # add minio info for storing git repo bundle
            add_minio = True
            bucket, folder = minioops.get_or_create_minio_bucket_for_git(step['checkout'], branch_id=agent.job.run.flow.branch_id)
            step['minio_bucket'] = bucket
            step['minio_folder'] = folder

        # custom fields for ARTIFACTS
        if step['tool'] == 'artifacts':
            add_minio = True
            step['minio_bucket'] = minio_bucket
            if 'destination' not in step:
                step['destination'] = '.'

        # custom fields for CACHE
        if step['tool'] == 'cache':
            add_minio = True
            bucket, folders = minioops.get_or_create_minio_bucket_for_cache(agent.job, step)
            step['minio_bucket'] = bucket
            if step['action'] == 'save':
                step['minio_folder'] = folders
            else:
                step['minio_folders'] = folders

        if step['tool_location'].startswith('minio:'):
            add_minio = True

        # add minio details
        if add_minio:
            step['minio_addr'] = minio_addr
            step['minio_access_key'] = minio_access_key
            step['minio_secret_key'] = minio_secret_key

    if not agent.job.started:
        agent.job.started = utils.utcnow()
        agent.job.state = consts.JOB_STATE_ASSIGNED
        db.session.commit()

    # prepare secrets
    secrets = dbutils.get_secret_values(agent.job.run.flow.branch.project)
    job['secrets'] = secrets

    return {'job': job}


def _store_results(job, step, result):
    t0 = time.time()
    q = TestCaseResult.query.filter_by(job=job)
    q = q.options(joinedload('test_case'))
    q = q.join('test_case')
    or_list = []
    results = {}
    for tr in result['test-results']:
        or_list.append(TestCase.name == tr['test'])
        results[tr['test']] = tr
        log.info('looking for existing results for TC: %s', tr)
    q = q.filter(or_(*or_list))

    # update status of existing test case results
    cnt = 0
    for tcr in q.all():
        log.info('updating result for %s', tcr.test_case.name)
        if tcr.test_case.name not in results:
            log.warning('MISSING result')
            continue
        tr = results.pop(tcr.test_case.name)
        tcr.cmd_line = tr['cmd']
        tcr.result = tr['status']
        tcr.values = tr['values'] if 'values' in tr else None
        cnt += 1
    db.session.commit()
    t1 = time.time()
    log.info('reporting %s existing test records took %ss', cnt, (t1 - t0))

    # create test case results if they didnt exist
    tool_test_cases = {}
    q = TestCase.query.filter_by(tool=step.tool)
    for tc in q.all():
        tool_test_cases[tc.name] = tc
    for tc_name, tr in results.items():
        tc = tool_test_cases.get(tc_name, None)
        if tc is None:
            tc = TestCase(name=tc_name, tool=step.tool)
        TestCaseResult(test_case=tc, job=step.job,
                       cmd_line=tr['cmd'],
                       result=tr['status'],
                       values=tr['values'] if 'values' in tr else None)
    db.session.commit()
    t2 = time.time()
    log.info('reporting %s new test records took %ss', len(results), (t2 - t1))


def _store_issues(job, result):
    t0 = time.time()
    for issue in result['issues']:
        issue_type = 0
        if issue['type'] in consts.ISSUE_TYPES_CODE:
            issue_type = consts.ISSUE_TYPES_CODE[issue['type']]
        else:
            log.warning('unknown issue type: %s', issue['type'])
        extra = {}
        for k, v in issue.items():
            if k not in ['line', 'column', 'path', 'symbol', 'message']:
                extra[k] = v
        Issue(issue_type=issue_type, line=issue['line'], column=issue['column'], path=issue['path'], symbol=issue['symbol'],
              message=issue['message'][:511], extra=extra, job=job)
    db.session.commit()
    t1 = time.time()
    log.info('reporting %s issues took %ss', len(result['issues']), (t1 - t0))


def _store_artifacts(job, step):
    t0 = time.time()
    flow = job.run.flow
    if not flow.artifacts:
        flow.artifacts = dict(public=dict(size=0, count=0, entries=[]),
                              private=dict(size=0, count=0))
    if not flow.artifacts_files:
        flow.artifacts_files = []

    run = job.run
    if not run.artifacts:
        run.artifacts = dict(public=dict(size=0, count=0, entries=[]),
                             private=dict(size=0, count=0))
    if not run.artifacts_files:
        run.artifacts_files = []

    public = step.fields.get('public', False)
    report_entry = step.fields.get('report_entry', None)
    if report_entry:
        public = True
        flow.artifacts['public']['entries'].append(report_entry)
        run.artifacts['public']['entries'].append(report_entry)

    if public:
        section = 'public'
        section_id = consts.ARTIFACTS_SECTION_PUBLIC
    else:
        section = 'private'
        section_id = consts.ARTIFACTS_SECTION_PRIVATE

    for artifact in step.result['artifacts']:
        flow.artifacts[section]['size'] += artifact['size']
        flow.artifacts[section]['count'] += 1

        run.artifacts[section]['size'] += artifact['size']
        run.artifacts[section]['count'] += 1

        if section == 'public':
            report_entry = artifact.get('report_entry', None)
            if report_entry:
                flow.artifacts['public']['entries'].append(report_entry)
                run.artifacts['public']['entries'].append(report_entry)

        path = artifact['path']
        f = File.query.filter_by(path=path).one_or_none()
        if f is None:
            f = File(path=path)
        Artifact(file=f, flow=flow, run=run, size=artifact['size'], section=section_id)

    flag_modified(flow, 'artifacts')
    flag_modified(run, 'artifacts')
    db.session.commit()

    t1 = time.time()
    log.info('reporting %s artifacts took %ss', len(step.result['artifacts']), (t1 - t0))


def destroy_machine_if_needed(agent, job):
    to_destroy = False

    # check if cloud machine should be destroyed now
    ag = dbutils.find_cloud_assignment_group(agent)
    if ag:
        # aws ec2
        if ag.deployment['method'] == consts.AGENT_DEPLOYMENT_METHOD_AWS_EC2:
            depl = ag.deployment['aws']

            if depl and 'destruction_after_jobs' in depl and int(depl['destruction_after_jobs']) > 0:
                max_jobs = int(depl['destruction_after_jobs'])
                q = Job.query.filter_by(agent_used=agent)
                q = q.filter(Job.finished.isnot(None))
                q = q.filter(Job.finished > agent.created)
                jobs_num = q.count()
                log.info('JOB %s, num %d, max %d', job, jobs_num, max_jobs)
                if jobs_num >= max_jobs:
                    to_destroy = True

        # aws ecs fargate
        elif ag.deployment['method'] == consts.AGENT_DEPLOYMENT_METHOD_AWS_ECS_FARGATE:
            log.info('ECS FARGATE JOB %s - destroying task: %d', job, agent.id)
            to_destroy = True

        # azure vm
        elif ag.deployment['method'] == consts.AGENT_DEPLOYMENT_METHOD_AZURE_VM:
            depl = ag.deployment['azure_vm']

            if depl and 'destruction_after_jobs' in depl and int(depl['destruction_after_jobs']) > 0:
                max_jobs = int(depl['destruction_after_jobs'])
                q = Job.query.filter_by(agent_used=agent)
                q = q.filter(Job.finished.isnot(None))
                q = q.filter(Job.finished > agent.created)
                jobs_num = q.count()
                log.info('JOB %s, num %d, max %d', job, jobs_num, max_jobs)
                if jobs_num >= max_jobs:
                    to_destroy = True

        # kubernetes
        elif ag.deployment['method'] == consts.AGENT_DEPLOYMENT_METHOD_K8S:
            log.info('K8S JOB %s - destroying pod: %d', job, agent.id)
            to_destroy = True

    # schedule destruction if needed
    if to_destroy:
        agent.disabled = True
        kkrq.enq(bg_jobs.destroy_machine, agent.id)


def _handle_step_result(agent, req):
    response = {}
    if agent.job is None:
        log.error('job in agent %s is missing, reporting some old job %s, step %s',
                  agent, req['job_id'], req['step_idx'])
        return response

    if agent.job_id != req['job_id']:
        log.error('agent %s is reporting some other job %s',
                  agent, req['job_id'])
        return response

    log.set_ctx(job=agent.job_id)

    try:
        result = req['result']
        step_idx = req['step_idx']
        status = result['status']
        del result['status']
        if status not in consts.STEP_STATUS_TO_INT:
            log.set_ctx(job=None)
            raise ValueError("unknown status: %s" % status)
    except Exception:
        log.exception('problems with parsing request')
        log.set_ctx(job=None)
        return response

    job = agent.job
    step = job.steps[step_idx]
    step.result = result
    step.status = consts.STEP_STATUS_TO_INT[status]

    # handle canceling situation
    if job.state == consts.JOB_STATE_COMPLETED:
        agent.job = None
        db.session.commit()
        log.info("canceling job %s on %s", job, agent)
        response['cancel'] = True
        log.set_ctx(job=None)
        return response

    db.session.commit()

    # store test results
    if 'test-results' in result:
        _store_results(job, step, result)

    # store issues
    if 'issues' in result:
        _store_issues(job, result)

    # store issues
    if 'artifacts' in result:
        _store_artifacts(job, step)

    # check if all steps are done so job is finised
    job_finished = True
    log.info('checking steps')
    for s in job.steps:
        log.info('%s: %s', s.index, consts.STEP_STATUS_NAME[s.status]
                 if s.status in consts.STEP_STATUS_NAME else s.status)
        if s.status == consts.STEP_STATUS_DONE:
            continue
        if s.status == consts.STEP_STATUS_ERROR:
            job_finished = True
            break
        job_finished = False
        break
    if job_finished:
        job.state = consts.JOB_STATE_EXECUTING_FINISHED
        job.finished = utils.utcnow()
        agent.job = None

        destroy_machine_if_needed(agent, job)

        db.session.commit()
        kkrq.enq(bg_jobs.job_completed, job.id)
        log.info('job %s finished by %s', job, agent)
    else:
        response['timeout'] = _left_time(job)

    log.set_ctx(job=None)
    return response


def _create_test_records(step, tests):
    t0 = time.time()
    tool_test_cases = {}
    q = TestCase.query.filter_by(tool=step.tool)
    for tc in q.all():
        tool_test_cases[tc.name] = tc

    for t in tests:
        tc = tool_test_cases.get(t, None)
        if tc is None:
            tc = TestCase(name=t, tool=step.tool)
        TestCaseResult(test_case=tc, job=step.job)
    db.session.commit()
    t1 = time.time()
    log.info('creating %s test records took %ss', len(tests), (t1 - t0))


def _handle_dispatch_tests(agent, req):
    if agent.job is None:
        log.error('job in agent %s is missing, reporting some old job %s, step %s',
                  agent, req['job_id'], req['step_idx'])
        return {}

    if agent.job_id != req['job_id']:
        log.error('agent %s is reporting some other job %s',
                  agent, req['job_id'])
        return {}

    job = agent.job

    # handle canceling situation
    if job.state == consts.JOB_STATE_COMPLETED:
        agent.job = None
        db.session.commit()
        log.info("canceling job %s on %s", job, agent)
        return {'cancel': True}

    try:
        tests = req['tests']
        step_idx = req['step_idx']
        step = job.steps[step_idx]
    except Exception:
        log.exception('problems with parsing request')
        return {}

    tests_cnt = len(tests)

    if len(set(tests)) != tests_cnt:
        log.warning('there are tests duplicates')
        return {}

    if tests_cnt == 0:
        # TODO
        raise NotImplementedError

    if tests_cnt == 1 or 'autosplit' not in step.fields or not step.fields['autosplit']:
        _create_test_records(step, tests)
        db.session.commit()
        return {'tests': tests}

    # simple dispatching: divide to 2 jobs, current and new one
    part = tests_cnt // 2
    part1 = tests[:part]
    part2 = tests[part:]

    _create_test_records(step, part1)
    db.session.commit()

    # new timeout reduced by nearly a half
    timeout = int(job.timeout * 0.6)
    timeout = max(timeout, 60)

    # create new job and its steps
    job2 = Job(run=job.run, name=job.name, agents_group=job.agents_group, system=job.system, timeout=timeout)
    for s in job.steps:
        s2 = Step(job=job2, index=s.index, tool=s.tool, fields=s.fields.copy())
        if s.index == step_idx:
            _create_test_records(s2, part2)
    db.session.commit()

    return {'tests': part1}


def _handle_host_info(agent, req):  # pylint: disable=unused-argument
    log.info('HOST INFO %s', req['info'])

    system = req['info']['system']

    sys = None
    if system.isdigit():
        # agent has already identified system so find its name
        sys_id = int(system)
        sys = System.query.filter_by(id=sys_id).one_or_none()
        if sys is not None:
            # if this is real system id then substitute it
            req['info']['system'] = sys.name

    agent.host_info = req['info']
    db.session.commit()

    resp = dict(agent_id=agent.id)

    if sys:
        # agent has already identified system so it doesn't need to be created
        return resp

    try:
        System(name=system, executor='local')
        db.session.commit()
    except IntegrityError as e:
        if not isinstance(e.orig, UniqueViolation):
            log.exception('IGNORED')

    return resp


def _handle_keep_alive(agent, req):  # pylint: disable=unused-argument
    job = agent.job

    # handle canceling situation
    if job and job.state == consts.JOB_STATE_COMPLETED:
        agent.job = None
        db.session.commit()
        log.info("canceling job %s on %s", job, agent)
        return {'cancel': True}

    return {}


def _handle_unknown_agent(address, ip_address, agent):
    try:
        new = False
        if agent:
            now = utils.utcnow()
            if agent.deleted:
                log.info('undeleting agent %s with address %s', agent, address)
                agent.deleted = None
                agent.created = now
            agent.authorized = False
            agent.ip_address = ip_address
            agent.last_seen = now
        else:
            agent = Agent(name=address, address=address, authorized=False, ip_address=ip_address, last_seen=utils.utcnow())
            new = True
        db.session.commit()
        if new:
            log.info('created new agent instance %s for address %s', agent, address)
    except Exception as e:
        log.warning('IGNORED EXCEPTION: %s', str(e))


def _serve_agent_request():
    log.reset_ctx()
    req = request.get_json()
    # log.info('request headers: %s', request.headers)
    # log.info('request remote_addr: %s', request.remote_addr)
    # log.info('request args: %s', request.args)
    log.info('request data: %s', str(req)[:200])

    msg = req['msg']
    address = req['address']
    if address is None:
        address = request.remote_addr
    # log.info('agent address: %s', address)

    agent = Agent.query.filter_by(address=address).one_or_none()
    if agent is None or agent.deleted:
        log.warning('unknown agent %s', address)
        _handle_unknown_agent(address, request.remote_addr, agent)
        return json.dumps({'unauthorized': True})

    agent.last_seen = utils.utcnow()
    db.session.commit()

    if not agent.authorized:
        log.warning('unauthorized agent %s from %s', address, request.remote_addr)
        return json.dumps({'unauthorized': True})

    response = {}

    if msg == consts.AGENT_MSG_GET_JOB:
        response = _handle_get_job(agent)

        clickhouse_addr = os.environ.get('KRAKEN_CLICKHOUSE_ADDR', consts.DEFAULT_CLICKHOUSE_ADDR)
        response['cfg'] = dict(clickhouse_addr=clickhouse_addr)
        response['version'] = version.version

    elif msg == consts.AGENT_MSG_STEP_RESULT:
        response = _handle_step_result(agent, req)

    elif msg == consts.AGENT_MSG_DISPATCH_TESTS:
        response = _handle_dispatch_tests(agent, req)

    elif msg == consts.AGENT_MSG_HOST_INFO:
        response = _handle_host_info(agent, req)

    elif msg == consts.AGENT_MSG_KEEP_ALIVE:
        response = _handle_keep_alive(agent, req)

    else:
        log.warning('unknown msg: %s', msg)
        response = {}

    log.info('sending response: %s', str(response)[:200])
    return json.dumps(response)


def serve_agent_request():
    try:
        return _serve_agent_request()
    except Exception:
        db.session.rollback()
        db.session.close()
        raise



"""add enabled to Stage

Revision ID: 0731897c862e
Revises:
Create Date: 2019-12-22 17:39:27.154403

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '0731897c862e'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('stages', sa.Column('enabled', sa.Boolean, default=True))


def downgrade():
    op.drop_column('stages', 'enabled')



"""added cancel to agent

Revision ID: 73db0f22e1e6
Revises: 89ffd52f1f00
Create Date: 2020-09-19 07:28:37.769815

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '73db0f22e1e6'
down_revision = '89ffd52f1f00'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('agents', sa.Column('cancel', sa.Boolean(), nullable=True))


def downgrade():
    op.drop_column('agents', 'cancel')



"""change project_id in ExecutorGroup to nullable

Revision ID: 735e81eedae4
Revises: 67d4f5a5ee98
Create Date: 2020-01-11 07:16:41.259814

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '735e81eedae4'
down_revision = '67d4f5a5ee98'
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column('executor_groups', 'project_id',
                    existing_type=sa.INTEGER(),
                    nullable=True)


def downgrade():
    op.alter_column('executor_groups', 'project_id',
                    existing_type=sa.INTEGER(),
                    nullable=False)



"""added schema repo to Stage

Revision ID: 6745adaf40e5
Revises: a7de4bdee425
Create Date: 2020-02-16 19:29:56.547139

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '6745adaf40e5'
down_revision = 'a7de4bdee425'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('stages', sa.Column('repo_access_token', sa.UnicodeText(), nullable=True))
    op.add_column('stages', sa.Column('repo_branch', sa.UnicodeText(), nullable=True))
    op.add_column('stages', sa.Column('repo_url', sa.UnicodeText(), nullable=True))
    op.add_column('stages', sa.Column('schema_file', sa.UnicodeText(), nullable=True))
    op.add_column('stages', sa.Column('schema_from_repo_enabled', sa.Boolean(), nullable=True))


def downgrade():
    op.drop_column('stages', 'schema_from_repo_enabled')
    op.drop_column('stages', 'schema_file')
    op.drop_column('stages', 'repo_url')
    op.drop_column('stages', 'repo_branch')
    op.drop_column('stages', 'repo_access_token')



"""added test case comment table

Revision ID: f566580fd139
Revises: a35aec7afa0a
Create Date: 2022-01-15 21:02:31.961891

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = 'f566580fd139'
down_revision = 'a35aec7afa0a'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table('test_case_comments',
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('test_case_id', sa.Integer(), nullable=True),
                    sa.Column('branch_id', sa.Integer(), nullable=True),
                    sa.Column('last_flow_id', sa.Integer(), nullable=True),
                    sa.Column('state', sa.Integer(), nullable=True),
                    sa.Column('data', postgresql.JSONB(astext_type=sa.Text()), nullable=True),
                    sa.ForeignKeyConstraint(['branch_id'], ['branches.id'], ),
                    sa.ForeignKeyConstraint(['last_flow_id'], ['flows.id'], ),
                    sa.ForeignKeyConstraint(['test_case_id'], ['test_cases.id'], ),
                    sa.PrimaryKeyConstraint('id'))
    op.add_column('test_case_results', sa.Column('comment_id', sa.Integer(), nullable=True))
    op.create_foreign_key('fk_test_case_comments_test_case_results',
                          'test_case_results', 'test_case_comments', ['comment_id'], ['id'])


def downgrade():
    op.drop_constraint('fk_test_case_comments_test_case_results', 'test_case_results', type_='foreignkey')
    op.drop_column('test_case_results', 'comment_id')
    op.drop_table('test_case_comments')



# Copyright 2021 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import random
import logging
import datetime

# AZURE
import azure
from azure.identity import ClientSecretCredential
from azure.mgmt.resource import ResourceManagementClient
from azure.mgmt.network import NetworkManagementClient
from azure.mgmt.network.models import SecurityRuleAccess, SecurityRuleDirection, SecurityRuleProtocol
from azure.mgmt.compute import ComputeManagementClient
from azure.mgmt.subscription import SubscriptionClient
from azure.mgmt.monitor import MonitorManagementClient

from .. import utils
from .. import dbutils
from ..models import get_setting
from .common import _create_agent


log = logging.getLogger(__name__)


# AZURE VM ####################################################################

def check_azure_settings():
    subscription_id = get_setting('cloud', 'azure_subscription_id')
    if not subscription_id:
        return 'Azure Subscription ID is empty'
    # TODO
    # if len(access_key) < 16:
    #     return 'AWS access key is too short'
    # if len(access_key) > 128:
    #     return 'AWS access key is too long'

    tenant_id = get_setting('cloud', 'azure_tenant_id')
    if not tenant_id:
        return 'Azure Tenant ID is empty'

    client_id = get_setting('cloud', 'azure_client_id')
    if not client_id:
        return 'Azure Client ID is empty'

    client_secret = get_setting('cloud', 'azure_client_secret')
    if not client_secret:
        return 'Azure Client Secret is empty'

    try:
        credential = ClientSecretCredential(tenant_id=tenant_id, client_id=client_id, client_secret=client_secret)
        subscription_client = SubscriptionClient(credential)
        subscription_client.subscriptions.list_locations(subscription_id)
    except Exception as ex:
        return str(ex)

    return 'ok'


def login_to_azure():
    subscription_id = get_setting('cloud', 'azure_subscription_id')
    tenant_id = get_setting('cloud', 'azure_tenant_id')
    client_id = get_setting('cloud', 'azure_client_id')
    client_secret = get_setting('cloud', 'azure_client_secret')
    settings = ['subscription_id', 'tenant_id', 'client_id', 'client_secret']
    for s in settings:
        val = locals()[s]
        if not val:
            log.error('Azure %s is empty, please set it in global cloud settings', s)
            return None, None

    credential = ClientSecretCredential(tenant_id=tenant_id, client_id=client_id, client_secret=client_secret)
    return credential, subscription_id


def _create_azure_vm(ag, system,
                     server_url, clickhouse_addr,
                     credential, subscription_id):
    instance_id = str(random.randrange(9999999999))

    _, depl = ag.get_deployment()

    location = depl['location']
    vm_size = depl['vm_size']

    sys_parts = system.name.split(':')
    if len(sys_parts) != 4:
        log.warning('incorrect system image name %s, Azure image name should have form <publisher>:<offer>:<sku>:<version>',
                    system.name)
        return

    log.info("Provisioning a virtual machine...some operations might take a minute or two.")

    # Step 1: Provision a resource group

    # Obtain the management object for resources, using the credential from the CLI login.
    resource_client = ResourceManagementClient(credential, subscription_id)

    # Provision the resource groups.
    global_rg = "kraken-rg"
    rg_result = resource_client.resource_groups.create_or_update(
        global_rg,
        {
            "location": location
        }
    )
    log.info("Provisioned global resource group %s in the %s region",
             rg_result.name, rg_result.location)

    group_rg = "kraken-%d-rg" % ag.id
    rg_result = resource_client.resource_groups.create_or_update(
        group_rg,
        {
            "location": location
        }
    )
    log.info("Provisioned resource group %s in the %s region",
             rg_result.name, rg_result.location)


    # For details on the previous code, see Example: Provision a resource group
    # at https://docs.microsoft.com/azure/developer/python/azure-sdk-example-resource-group


    # Step 2: provision a virtual network

    # A virtual machine requires a network interface client (NIC). A NIC requires
    # a virtual network and subnet along with an IP address. Therefore we must provision
    # these downstream components first, then provision the NIC, after which we
    # can provision the VM.

    # Network and IP address names
    vnet_name = "kraken-vnet"
    subnet_name = "kraken-subnet"

    ip_config_name = "kraken-%s-ip-config" % instance_id
    ip_name = "kraken-%s-ip" % instance_id
    nic_name = "kraken-%s-nic" % instance_id
    vm_name = "kraken-agent-%s-vm" % instance_id

    # Obtain the management object for networks
    network_client = NetworkManagementClient(credential, subscription_id)

    # if VNET exists then use it, otherwise create new one
    try:
        vnet_result = network_client.virtual_networks.get(global_rg, vnet_name)
    except azure.core.exceptions.ResourceNotFoundError:
        # Provision the virtual network and wait for completion
        poller = network_client.virtual_networks.begin_create_or_update(
            global_rg,
            vnet_name,
            {
                "location": location,
                "address_space": {
                    "address_prefixes": ["10.0.0.0/16"]
                }
            })

        vnet_result = poller.result()

    log.info("Provisioned virtual network %s with address prefixes %s",
             vnet_result.name, vnet_result.address_space.address_prefixes)

    #########
    security_group_name = 'kraken-nsg'
    try:
        security_group = network_client.network_security_groups.get(global_rg, security_group_name)
    except azure.core.exceptions.ResourceNotFoundError:
        #security_group_params = NetworkSecurityGroup(
        #    id="testnsg",
        #    location=self.location,
        #    tags={"name": security_group_name}
        #)
        poller = network_client.network_security_groups.begin_create_or_update(
            global_rg,
            security_group_name,
            {
                "location": location,
            }
        )
        security_group = poller.result()

    ssh_security_rule_name = 'kraken-ssh-rule'
    try:
        network_client.security_rules.get(global_rg, security_group_name, ssh_security_rule_name)
    except azure.core.exceptions.ResourceNotFoundError:
        poller = network_client.security_rules.begin_create_or_update(
            global_rg,
            security_group_name,
            ssh_security_rule_name,
            {
                'access': SecurityRuleAccess.allow,
                'description': 'SSH security rule',
                'destination_address_prefix': '*',
                'destination_port_range': '22',
                'direction': SecurityRuleDirection.inbound,
                'priority': 400,
                'protocol': SecurityRuleProtocol.tcp,
                'source_address_prefix': '*',  # TODO: change it to my_ip
                'source_port_range': '*',
            }
        )
        poller.result()
    #########

    # Step 3: Provision the subnet and wait for completion
    # if subnet exists then use it, otherwise create new one
    try:
        subnet_result = network_client.subnets.get(global_rg, vnet_name, subnet_name)
    except Exception:
        poller = network_client.subnets.begin_create_or_update(
            global_rg,
            vnet_name,
            subnet_name,
            {
                "address_prefix": "10.0.0.0/24",
            }
        )
        subnet_result = poller.result()

    log.info("Provisioned virtual subnet %s with address prefix %s",
             subnet_result.name, subnet_result.address_prefix)

    # Step 4: Provision an IP address and wait for completion
    poller = network_client.public_ip_addresses.begin_create_or_update(
        group_rg,
        ip_name,
        {
            "location": location,
            "sku": { "name": "Standard" },
            "public_ip_allocation_method": "Static",
            "public_ip_address_version" : "IPV4"
        }
    )
    ip_address_result = poller.result()

    log.info("Provisioned public IP address %s with address %s",
             ip_address_result.name, ip_address_result.ip_address)

    # Step 5: Provision the network interface client
    poller = network_client.network_interfaces.begin_create_or_update(
        group_rg,
        nic_name,
        {
            "location": location,
            "ip_configurations": [{
                "name": ip_config_name,
                "subnet": { "id": subnet_result.id },
                "public_ip_address": {"id": ip_address_result.id }
            }],
            "network_security_group": {"id": security_group.id }
        }
    )

    nic_result = poller.result()

    log.info("Provisioned network interface client %s", nic_result.name)

    # Step 6: Provision the virtual machine

    # Obtain the management object for virtual machines
    compute_client = ComputeManagementClient(credential, subscription_id)

    username = "kraken"
    password = "kraken123!"

    log.info("Provisioning virtual machine %s; this operation might take a few minutes.", vm_name)

    # Provision the VM specifying only minimal arguments, which defaults to an Ubuntu 18.04 VM
    # on a Standard DS1 v2 plan with a public IP address and a default virtual network/subnet.

    init_script = """#!/usr/bin/env bash
                     exec > >(tee /var/log/user-data.log|logger -t user-data -s 2>/dev/console) 2>&1
                     wget -O agent {server_url}/bk/install/agent
                     chmod a+x agent
                     ./agent install -s {server_url} -c {clickhouse_addr} --system-id {system_id}
                  """
    init_script = init_script.format(server_url=server_url, clickhouse_addr=clickhouse_addr,
                                     system_id=system.id)

    poller = compute_client.virtual_machines.begin_create_or_update(
        group_rg,
        vm_name,
        {
            "location": location,
            "storage_profile": {
                "image_reference": {
                    "publisher": sys_parts[0],  # 'Canonical',
                    "offer": sys_parts[1],  # "0001-com-ubuntu-server-focal",
                    "sku": sys_parts[2],  # "20_04-lts",
                    "version": sys_parts[3]  # "latest"
                }
                #"image_reference": "Canonical:0001-com-ubuntu-server-focal:20_04-lts:20.04.202109080"
            },
            "hardware_profile": {
                "vm_size": vm_size
            },
            "os_profile": {
                "computer_name": vm_name,
                "admin_username": username,
                "admin_password": password,
            },
            "network_profile": {
                "network_interfaces": [{
                    "id": nic_result.id,
                }]
            },
            "user_data": base64.b64encode(init_script.encode('utf-8')).decode()
        }
    )

    vm_result = poller.result()

    log.info("Provisioned Azure virtual machine %s", vm_result.name)

    # VM params
    sys_id = system.id if system.executor == 'local' else 0
    address = nic_result.ip_configurations[0].private_ip_address
    params = dict(name=vm_name,
                  address=address,
                  ip_address=ip_address_result.ip_address,
                  extra_attrs=dict(
                      system=sys_id,
                      instance_id=instance_id))

    a = _create_agent(params, ag)
    log.info('spawned new agent %s on Azure VM instance %s at %s', a, vm_result.name, a.created)


def create_vms(ag, system, num,
               server_url, clickhouse_addr):

    # Acquire a credential object using service principal authentication.
    credential, subscription_id = login_to_azure()
    if not credential:
        return

    for _ in range(num):
        _create_azure_vm(ag, system,
                         server_url, clickhouse_addr,
                         credential, subscription_id)


def _destroy_azure_vm(rg, vm_name, cc, nc):
    # get instance id from vm name, name is as follows "kraken-agent-<instance_id>-vm"
    instance_id = vm_name[13:-3]

    try:
        vm = cc.virtual_machines.get(rg, vm_name)
        disk_name = vm.storage_profile.os_disk.name
        cc.virtual_machines.begin_delete(rg, vm_name).wait()
        cc.disks.begin_delete(rg, disk_name)
    except azure.core.exceptions.ResourceNotFoundError:
        log.info('azure vm %s already missing', vm_name)

    nic_name = "kraken-%s-nic" % instance_id
    try:
        nc.network_interfaces.begin_delete(rg, nic_name).wait()
    except azure.core.exceptions.ResourceNotFoundError:
        log.info('azure nic %s already missing', nic_name)

    # ip_config_name = "kraken-%s-ip-config" % instance_id
    ip_name = "kraken-%s-ip" % instance_id
    try:
        nc.public_ip_addresses.begin_delete(rg, ip_name).wait()
    except azure.core.exceptions.ResourceNotFoundError:
        log.info('azure ip %s already missing', ip_name)

    log.info('deleted azure vm: %s', vm_name)


def destroy_vm(ag, agent):  # pylint: disable=unused-argument
    instance_id = agent.extra_attrs['instance_id']
    vm_name = "kraken-agent-%s-vm" % instance_id
    rg = 'kraken-%d-rg' % ag.id

    log.info('deleting azure vm: %s', vm_name)

    # Acquire a credential object using service principal authentication.
    credential, subscription_id = login_to_azure()
    if not credential:
        return

    cc = ComputeManagementClient(credential, subscription_id)
    nc = NetworkManagementClient(credential, subscription_id)

    _destroy_azure_vm(rg, vm_name, cc, nc)


def vm_exists(ag, agent):
    instance_id = agent.extra_attrs['instance_id']
    ag = dbutils.find_cloud_assignment_group(agent)
    if not ag:
        return False
    rg = 'kraken-%d-rg' % ag.id
    vm_name = "kraken-agent-%s-vm" % instance_id

    credential, subscription_id = login_to_azure()
    if not credential:
        raise Exception('wrong azure credential')

    cc = ComputeManagementClient(credential, subscription_id)
    try:
        cc.virtual_machines.get(rg, vm_name)
    except azure.core.exceptions.ResourceNotFoundError:
        return False

    return True


def cleanup_dangling_vms(ag):  # pylint: disable=unused-argument
    credential, subscription_id = login_to_azure()
    if not credential:
        return 0, 0, 0, 0, 0

    cc = ComputeManagementClient(credential, subscription_id)
    mc = MonitorManagementClient(credential, subscription_id)
    nc = NetworkManagementClient(credential, subscription_id)

    rg = 'kraken-%d-rg' % ag.id
    now = utils.utcnow()

    try:
        vms = cc.virtual_machines.list(rg)
    except azure.core.exceptions.ResourceNotFoundError:
        return 0, 0, 0, 0, 0

    instances = 0
    terminated_instances = 0
    assigned_instances = 0
    orphaned_instances = 0
    orphaned_terminated_instances = 0

    for vm in vms:
        instances += 1

        # if vm is being deleted then skip it
        vm2 = cc.virtual_machines.instance_view(rg, vm.name)
        log.info('vm %s statuses:', vm.name)
        skip = False
        for s in vm2.statuses:
            log.info(s.code)
            if s.code == 'ProvisioningState/deleting':
                skip = True
                break
        if skip:
            continue

        # if assigned to some agent then skip it
        assigned = False
        for aa in ag.agents:
            agent = aa.agent
            if agent.extra_attrs and 'instance_id' in agent.extra_attrs:
                instance_id = agent.extra_attrs['instance_id']
                vm_name = "kraken-agent-%s-vm" % instance_id
                if vm_name == vm.name:
                    assigned = True
                    break
        if assigned:
            assigned_instances += 1
            continue

        # instances have to be old enough to avoid race condition with
        # case when instances are being created but not yet assigned to agents
        fltr = " and ".join([ "eventTimestamp ge '{}T00:00:00Z'".format(now.date()),
                              "resourceUri eq '%s'" % vm.id ])
        created_at = None
        for l in mc.activity_logs.list(filter=fltr, select="eventTimestamp"):
            created_at = l.event_timestamp
        if not created_at or now - created_at < datetime.timedelta(minutes=10):
            continue

        # the instance is not terminated, not assigned, old enough
        # so delete it as it seems to be a lost instance
        log.info('terminating lost azure vm instance %s', vm.name)
        orphaned_instances += 1
        try:
            _destroy_azure_vm(rg, vm.name, cc, nc)
        except Exception:
            log.exception('IGNORED EXCEPTION')

        orphaned_terminated_instances += 1

    return instances, terminated_instances, assigned_instances, orphaned_instances, orphaned_terminated_instances



"""added several stats fields to Run

Revision ID: 8200318b9e18
Revises: b908e219f07f
Create Date: 2020-02-08 11:50:24.255977

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '8200318b9e18'
down_revision = 'b908e219f07f'
branch_labels = None
depends_on = None


def upgrade():
    op.drop_table('preferences')
    op.add_column('runs', sa.Column('issues_new', sa.Integer(), nullable=True))
    op.add_column('runs', sa.Column('issues_total', sa.Integer(), nullable=True))
    op.add_column('runs', sa.Column('jobs_error', sa.Integer(), nullable=True))
    op.add_column('runs', sa.Column('jobs_total', sa.Integer(), nullable=True))
    op.add_column('runs', sa.Column('tests_not_run', sa.Integer(), nullable=True))
    op.add_column('runs', sa.Column('tests_passed', sa.Integer(), nullable=True))
    op.add_column('runs', sa.Column('tests_total', sa.Integer(), nullable=True))


def downgrade():
    op.drop_column('runs', 'tests_total')
    op.drop_column('runs', 'tests_passed')
    op.drop_column('runs', 'tests_not_run')
    op.drop_column('runs', 'jobs_total')
    op.drop_column('runs', 'jobs_error')
    op.drop_column('runs', 'issues_total')
    op.drop_column('runs', 'issues_new')
    op.create_table('preferences',
                    sa.Column('id', sa.INTEGER(), autoincrement=True, nullable=False),
                    sa.Column('name', sa.VARCHAR(length=50), autoincrement=False, nullable=True),
                    sa.Column('value', sa.TEXT(), autoincrement=False, nullable=True),
                    sa.Column('val_type', sa.VARCHAR(length=8), autoincrement=False, nullable=True),
                    sa.PrimaryKeyConstraint('id', name='preferences_pkey'))



#!/usr/bin/env python3

# Copyright 2020 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
import mimetypes
from queue import Queue, Empty
from threading import Thread, Event

from flask import abort, Response
import minio

from .models import Run, Flow
from . import users
from . import access
from . import consts


log = logging.getLogger('storage')


class MinioDownloader:
    def __init__(self, bucket_name, timeout=0.01):
        minio_addr = os.environ.get('KRAKEN_MINIO_ADDR', consts.DEFAULT_MINIO_ADDR)
        root_user = os.environ['MINIO_ROOT_USER']
        root_password = os.environ['MINIO_ROOT_PASSWORD']

        self.mc = minio.Minio(minio_addr, access_key=root_user, secret_key=root_password, secure=False)
        found = self.mc.bucket_exists(bucket_name)
        if not found:
            raise Exception('missing %s minio bucket' % bucket_name)

        self.bucket_name = bucket_name

        self.timeout = timeout

        self.bytes = Queue()
        self.finished = Event()
        self.worker = None

    def get_bytes(self, filename):
        resp = self.mc.get_object(self.bucket_name, filename)
        for chunk in resp.stream():
            self.bytes.put(chunk)
        resp.release_conn()
        self.bytes.join()   # wait for all blocks in the queue to be marked as processed
        self.finished.set() # mark streaming as finished

    def send_bytes(self):
        while not self.finished.is_set():
            try:
                yield self.bytes.get(timeout=self.timeout)
                self.bytes.task_done()
            except Empty:
                self.finished.wait(self.timeout)
        self.worker.join()

    def download(self, filename):
        self.worker = Thread(target=self.get_bytes, args=(filename,))
        self.worker.start()
        return self.send_bytes()


def serve_artifact(store_type, flow_id, run_id, path):
    log.info('path %s, %s, %s, %s', store_type, flow_id, run_id, path)

    if store_type not in ['public', 'report']:
        abort(400, "Not supported store type: %s" % store_type)

    if flow_id:
        flow = Flow.query.filter_by(id=int(flow_id)).one_or_none()
        if flow is None:
            abort(404, "Flow not found")

        runs = []
        for r in flow.runs:
            runs.append(r.id)
        runs.sort()
        run_id = runs[-1]
    else:
        run = Run.query.filter_by(id=int(run_id)).one_or_none()
        if run is None:
            abort(404, "Run not found")
        flow = run.flow


    token_info = users.get_token_info_from_request()
    if not token_info:
        abort(401, "Missing Authorization header or kk_session_token cookie")
    access.check(token_info, flow.branch.project_id, 'view',
                 'only superadmin, project admin, project power user and project viewer roles can fetch artifacts')

    mt, _ = mimetypes.guess_type(path)
    if mt is None:
        mt = 'application/octet-stream'

    bucket_name = '%08d' % flow.branch_id
    mc_dl = MinioDownloader(bucket_name)

    path = os.path.join(str(flow.id), str(run_id), path)
    resp = Response(mc_dl.download(path), mimetype=mt)

    return resp


def serve_flow_artifact(store_type, flow_id, path):
    return serve_artifact(store_type, flow_id, None, path)


def serve_run_artifact(store_type, run_id, path):
    return serve_artifact(store_type, None, run_id, path)



"""added deployment to agents group

Revision ID: 050fbc3d126c
Revises: 26c6d691e7ce
Create Date: 2021-05-03 11:31:59.682708

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '050fbc3d126c'
down_revision = '26c6d691e7ce'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('agents_groups', sa.Column('deployment', postgresql.JSONB(astext_type=sa.Text()), nullable=True))


def downgrade():
    op.drop_column('agents_groups', 'deployment')



"""removed cancel from agent

Revision ID: 38088514157a
Revises: 73db0f22e1e6
Create Date: 2020-09-20 06:56:12.245924

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '38088514157a'
down_revision = '73db0f22e1e6'
branch_labels = None
depends_on = None


def upgrade():
    op.drop_column('agents', 'cancel')


def downgrade():
    op.add_column('agents', sa.Column('cancel', sa.BOOLEAN(), autoincrement=False, nullable=True))



"""added last flows to branch

Revision ID: 26c6d691e7ce
Revises: 5b13b262b132
Create Date: 2021-04-10 23:15:08.698459

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '26c6d691e7ce'
down_revision = '5b13b262b132'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('branches', sa.Column('ci_last_completed_flow_id', sa.Integer(), nullable=True))
    op.add_column('branches', sa.Column('ci_last_incomplete_flow_id', sa.Integer(), nullable=True))
    op.create_foreign_key('branches_ci_last_incomplete_flow_id_fkey', 'branches', 'flows', ['ci_last_incomplete_flow_id'], ['id'])
    op.create_foreign_key('branches_ci_last_completed_flow_id_fkey', 'branches', 'flows', ['ci_last_completed_flow_id'], ['id'])


def downgrade():
    op.drop_constraint('branches_ci_last_incomplete_flow_id_fkey', 'branches', type_='foreignkey')
    op.drop_constraint('branches_ci_last_completed_flow_id_fkey', 'branches', type_='foreignkey')
    op.drop_column('branches', 'ci_last_incomplete_flow_id')
    op.drop_column('branches', 'ci_last_completed_flow_id')



"""added timeouts to Stage and Job

Revision ID: 3c1af14e0370
Revises: ff4a425a6070
Create Date: 2020-01-14 07:26:17.893236

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '3c1af14e0370'
down_revision = 'ff4a425a6070'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('jobs', sa.Column('timeout', sa.Integer(), nullable=True))
    op.add_column('stages', sa.Column('timeouts', postgresql.JSONB(astext_type=sa.Text()), nullable=True))


def downgrade():
    op.drop_column('stages', 'timeouts')
    op.drop_column('jobs', 'timeout')



"""move trigger_data to Flow

Revision ID: d45087b62027
Revises: b5cc62f79980
Create Date: 2019-12-27 07:24:54.315303

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = 'd45087b62027'
down_revision = 'b5cc62f79980'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('flows', sa.Column('trigger_data', postgresql.JSONB(astext_type=sa.Text()), nullable=True))
    op.drop_column('runs', 'trigger_data')


def downgrade():
    op.add_column('runs', sa.Column('trigger_data', postgresql.JSONB(astext_type=sa.Text()), autoincrement=False, nullable=True))
    op.drop_column('flows', 'trigger_data')



# Copyright 2020-2022 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
import logging

from flask import abort, request
from passlib.hash import pbkdf2_sha256
from werkzeug.exceptions import Unauthorized, BadRequest, NotFound, Forbidden
from sqlalchemy.sql.expression import asc, desc
from sqlalchemy.orm.attributes import flag_modified

from .models import db, User, UserSession, Project
from . import utils
from . import access
from . import authn

log = logging.getLogger(__name__)


def _check_user_password(user_id_or_name, password):
    # find user for given name
    q = User.query
    if isinstance(user_id_or_name, str):
        q = q.filter_by(name=user_id_or_name)
    else:
        q = q.filter_by(id=user_id_or_name)
    user = q.filter_by(deleted=None).one_or_none()

    if user:
        # check password
        if not user.password:
            if not user.details:
                return None
            ldap_user = authn.authenticate_ldap(user_id_or_name, password)
            if not ldap_user:
                return None
        elif not pbkdf2_sha256.verify(password, user.password):
            return None
    else:
        ldap_user = authn.authenticate_ldap(user_id_or_name, password)
        if ldap_user:
            details = dict(idp_type='ldap', idp='ldap://ldap.forumsys.com')
            if ldap_user['email']:
                details['email'] = ldap_user['email']
            user = User(name=ldap_user['username'], password='', details=details)
            db.session.commit()

    return user


def login(body):
    if 'method' not in body:
        method = 'local'
    else:
        method = body['method']

    user = None
    details = {}
    redirect_url = None

    if method == 'local':
        if 'user' not in body or 'password' not in body:
            raise BadRequest('bad credentials')

        # find user for given name with given password
        user = _check_user_password(body['user'], body['password'])
        if user is None:
            raise Unauthorized('missing user or incorrect password')
        token = uuid.uuid4().hex

        if user.details and not user.details.get('enabled', True):
            raise Unauthorized('user account is disabled')

    elif method == 'oidc':
        if 'oidc_provider' not in body:
            raise BadRequest('missing OIDC provider')
        if body['oidc_provider'] not in ['google', 'microsoft', 'github', 'auth0']:
            raise BadRequest('bad OIDC provider: %s' % body['oidc_provider'])

        idp = body['oidc_provider']
        redirect_url, state_data = authn.authenticate_oidc(idp)
        token = state_data['state']
        details['idp_type'] = 'oidc'
        details['idp'] = idp
        details['state'] = state_data

    # prepare user session
    us = UserSession(user=user, token=token, details=details)
    db.session.commit()

    resp = us.get_json()

    if user:
        roles_data = access.get_user_roles(user)
        resp['roles'] = roles_data

    if redirect_url:
        resp['redirect_url'] = redirect_url

    return resp, 201


def get_session(session_token, token_info=None):
    if token_info['session'].token != session_token:
        raise Forbidden('only user can get its session data')

    us = UserSession.query.filter_by(token=session_token, deleted=None).one_or_none()
    if us is None:
        raise BadRequest('user session not found')

    resp = us.get_json()
    resp['roles'] = access.get_user_roles(us.user)

    return resp, 200


def logout(session_token, token_info=None):
    if token_info['session'].token != session_token:
        raise Forbidden('only user can logout itself')

    us = UserSession.query.filter_by(token=session_token, deleted=None).one_or_none()
    if us is None:
        return None
    us.deleted = utils.utcnow()
    db.session.commit()

    if us.details and 'idp_type' in us.details and us.details['idp_type'] == 'oidc':
        authn.oidc_logout(us.details)

    return None


def check_auth_token(token):
    us = UserSession.query.filter_by(token=token, deleted=None).one_or_none()
    if us is None:
        raise Unauthorized
    resp = dict(sub=us.user, session=us)
    return resp


def get_token_info_from_request():
    token = request.headers.get('Authorization', None)
    if token:
        token = token.replace('Bearer ', '')
    else:
        token = request.cookies.get('kk_session_token')
    if not token:
        return None
    token_info = check_auth_token(token)
    return token_info


def change_password(user_id, body, token_info=None):
    if token_info['sub'].id != user_id:
        access.check(token_info, '', 'admin', 'only superadmin role can change users passwords')

    user_password = body
    if 'password_new' not in user_password or not user_password['password_new']:
        raise BadRequest('new password cannot be empty')

    user = _check_user_password(int(user_id), user_password['password_old'])

    if user is None:
        raise BadRequest('user not found or old password is incorrect')

    user.password = pbkdf2_sha256.hash(user_password['password_new'])
    db.session.commit()


def create_user(body, token_info=None):
    access.check(token_info, '', 'admin', 'only superadmin role can create users')

    for f in ['name', 'password']:
        if f not in body:
            abort(400, "Missing %s in user" % f)
        if  not body[f]:
            abort(400, "Empty %s in user" % f)

    user = User.query.filter_by(name=body['name']).one_or_none()
    if user is not None:
        abort(400, "User with name %s already exists" % body['name'])

    new_user = User(name=body['name'], password=pbkdf2_sha256.hash(body['password']))
    db.session.commit()

    return new_user.get_json(), 201


def get_users(start=0, limit=30, sort_field="name", sort_dir="asc", token_info=None):
    access.check(token_info, '', 'view', 'only superadmin role can get users')

    q = User.query
    q = q.filter_by(deleted=None)
    q = q.order_by(User.name)

    total = q.count()

    sort_func = asc
    if sort_dir == "desc":
        sort_func = desc

    if sort_field in ['name', 'id']:
        q = q.order_by(sort_func(sort_field))

    q = q.offset(start).limit(limit)

    users = []
    for u in q.all():
        users.append(u.get_json())
    return {'items': users, 'total': total}, 200


def get_user(user_id, token_info=None):
    access.check(token_info, '', 'view', 'only superadmin role can get user')

    q = User.query
    q = q.filter_by(id=user_id)
    q = q.filter_by(deleted=None)
    user = q.one_or_none()

    if user is None:
        raise NotFound('user not found')

    user_data = user.get_json()

    roles_data = access.get_user_roles(user)
    user_data.update(roles_data)

    return user_data, 200


def change_user_details(user_id, body, token_info=None):
    access.check(token_info, '', 'admin', 'only superadmin role can create users')

    q = User.query
    q = q.filter_by(id=user_id)
    q = q.filter_by(deleted=None)
    user = q.one_or_none()

    if user is None:
        raise NotFound('user %s not found' % user_id)

    if not user.details:
        user.details = {}

    if 'enabled' in body:
        if body['enabled']:
            user.details['enabled'] = True
        else:
            user.details['enabled'] = False
        flag_modified(user, 'details')
        db.session.commit()

    enforcer_reloaded = False
    policy_changed = False
    if 'superadmin' in body:
        policy_changed = True
        if not enforcer_reloaded:
            access.enforcer.load_policy()
            enforcer_reloaded = True

        if body['superadmin']:
            access.enforcer.add_named_grouping_policy("g2", str(user.id), access.ROLE_SUPERADMIN)
        else:
            access.enforcer.remove_named_grouping_policy("g2", str(user.id), access.ROLE_SUPERADMIN)
        db.session.commit()

    if 'projects' in body:
        policy_changed = True
        if not enforcer_reloaded:
            access.enforcer.load_policy()
            enforcer_reloaded = True

        for proj_id, role in body['projects'].items():
            q = Project.query
            q = q.filter_by(id=proj_id)
            q = q.filter_by(deleted=None)
            proj = q.one_or_none()
            if proj is None:
                raise NotFound('project %s not found' % proj_id)
            if role not in [access.ROLE_VIEWER, access.ROLE_PWRUSR, access.ROLE_ADMIN, None]:
                raise BadRequest('role %s not supported' % role)

            proj_role_viewer = 'viewer-p%d' % proj.id
            proj_role_pwrusr = 'pwrusr-p%d' % proj.id
            proj_role_admin = 'admin-p%d' % proj.id

            for r in [proj_role_viewer, proj_role_pwrusr, proj_role_admin]:
                access.enforcer.remove_named_grouping_policy("g", str(user.id), r)

            if role == access.ROLE_VIEWER:
                # p, viewer-p1, proj1, view
                access.enforcer.add_policy(proj_role_viewer, str(proj.id), 'view')
                # g, user, viewer-p1
                access.enforcer.add_named_grouping_policy("g", str(user.id), proj_role_viewer)
            elif role == access.ROLE_PWRUSR:
                # p, pwrusr-p1, proj1, view
                access.enforcer.add_policy(proj_role_pwrusr, str(proj.id), 'view')
                # p, pwrusr-p1, proj1, pwrusr
                access.enforcer.add_policy(proj_role_pwrusr, str(proj.id), 'pwrusr')
                # g, user, pwrusr-p1
                access.enforcer.add_named_grouping_policy("g", str(user.id), proj_role_pwrusr)
            elif role == access.ROLE_ADMIN:
                # p, admin-p1, proj1, view
                access.enforcer.add_policy(proj_role_admin, str(proj.id), 'view')
                # p, admin-p1, proj1, pwrusr
                access.enforcer.add_policy(proj_role_admin, str(proj.id), 'pwrusr')
                # p, admin-p1, proj1, admin
                access.enforcer.add_policy(proj_role_admin, str(proj.id), 'admin')
                # g, user, admin-p1
                access.enforcer.add_named_grouping_policy("g", str(user.id), proj_role_admin)
            db.session.commit()

    if policy_changed:
        access.notify_policy_change()

    user_data = user.get_json()
    roles_data = access.get_user_roles(user)
    user_data.update(roles_data)

    return user_data, 201



"""added system to Job

Revision ID: 1982be95e79f
Revises: 3c1af14e0370
Create Date: 2020-01-18 08:09:39.372105

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '1982be95e79f'
down_revision = '3c1af14e0370'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('jobs', sa.Column('system', sa.Unicode(length=200), nullable=True))


def downgrade():
    op.drop_column('jobs', 'system')



"""added index to flows

Revision ID: ab9326d5fc02
Revises: 9d8d8713c1ad
Create Date: 2021-05-28 07:44:40.012276

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = 'ab9326d5fc02'
down_revision = '9d8d8713c1ad'
branch_labels = None
depends_on = None


INDEXES = [
    # create INDEX ix_flows_branch_id_kind on flows(branch_id, kind);
    ['ix_flows_branch_id_kind', 'flows', 'branch_id, kind'],
    # create INDEX ix_flows_created on flows(created);
    ['ix_flows_created', 'flows', 'created'],
    # create INDEX ix_runs_flow_id on runs(flow_id);
    ['ix_runs_flow_id', 'runs', 'flow_id'],
    # create INDEX ix_artifacts_run_id on artifacts(run_id);
    ['ix_artifacts_run_id', 'artifacts', 'run_id'],
    # create INDEX ix_artifacts_flow_id on artifacts(flow_id);
    ['ix_artifacts_flow_id', 'artifacts', 'flow_id'],
]


def upgrade():
    conn = op.get_bind()
    for name, table, columns in INDEXES:
        print('creating index %s' % name)
        cmd = "CREATE INDEX IF NOT EXISTS %s ON %s (%s);" % (name, table, columns)
        conn.execute(cmd)
        #op.create_index(name, table, columns)
    print('migration completed')


def downgrade():
    conn = op.get_bind()
    for name, _, _ in INDEXES:
        print('dropping index %s' % name)
        cmd = "DROP INDEX IF EXISTS %s;" % name
        conn.execute(cmd)
        #op.drop_index(name, table_name=table)
    print('migration completed')



import os
import alembic.config
from flask import Flask
import sqlalchemy
import psycopg2.errors

from kraken.server import models, initdb, consts, srvcheck

def _get_db_version(app):
    with app.app_context():
        db_version = None
        try:
            db_version = models.AlembicVersion.query.one_or_none()
            if db_version is not None:
                db_version = db_version.version_num
        except sqlalchemy.exc.ProgrammingError as ex:
            if not isinstance(ex.orig, psycopg2.errors.UndefinedTable):
                raise

    return db_version


def main():
    print('Kraken DB Migration')

    # Create  Flask app instance
    app = Flask('Kraken DB Migration')

    db_url = os.environ.get('KRAKEN_DB_URL', consts.DEFAULT_DB_URL)
    print('migrating db: %s' % db_url)

    # wait for db
    srvcheck.check_postgresql(db_url)

    # Configure the SqlAlchemy part of the app instance
    app.config["SQLALCHEMY_ECHO"] = False
    app.config["SQLALCHEMY_DATABASE_URI"] = db_url
    app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False

    # initialize SqlAlchemy
    models.db.init_app(app)

    db_version = _get_db_version(app)

    here = os.path.dirname(os.path.abspath(__file__))
    alembic_ini_path = os.path.join(here, 'alembic.ini')
    alembic_args = [
        '-c', alembic_ini_path,
    ]

    if db_version is None:
        # create tables in db
        models.db.create_all(app=app)

        # stamp the alembic version of db schema in db
        alembic_args.extend(['stamp', 'head'])
        alembic.config.main(argv=alembic_args)
        print('alembic stamp completed')
    else:
        alembic_args.extend(['upgrade', 'head'])
        alembic.config.main(argv=alembic_args)
        print('alembic migration completed')

    with app.app_context():
        initdb.prepare_initial_data()

    db_version = _get_db_version(app)
    print('DB version: %s' % str(db_version))
    print('Kraken DB Migration completed')


if __name__ == "__main__":
    main()



"""added repo changes

Revision ID: 5b13b262b132
Revises: 770cfcc80cfb
Create Date: 2021-04-07 06:02:59.604892

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.sql import insert
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '5b13b262b132'
down_revision = '770cfcc80cfb'
branch_labels = None
depends_on = None


def upgrade():
    conn = op.get_bind()
    orm.Session(bind=conn)

    rc_tbl = op.create_table('repo_changes',
                             sa.Column('created', sa.DateTime(), nullable=False),
                             sa.Column('updated', sa.DateTime(), nullable=False),
                             sa.Column('deleted', sa.DateTime(), nullable=True),
                             sa.Column('id', sa.Integer(), nullable=False),
                             sa.Column('data', postgresql.JSONB(astext_type=sa.Text()), nullable=False),
                             sa.PrimaryKeyConstraint('id')
    )

    op.add_column('flows', sa.Column('trigger_data_id', sa.Integer()))
    op.create_foreign_key('flows_trigger_data_id_fkey', 'flows', 'repo_changes', ['trigger_data_id'], ['id'])

    op.add_column('runs', sa.Column('repo_data_id', sa.Integer()))
    op.create_foreign_key('runs_repo_data_id_fkey', 'runs', 'repo_changes', ['repo_data_id'], ['id'])

    print('updating flows')
    res = conn.execute("SELECT id, created, trigger_data FROM flows WHERE trigger_data IS NOT NULL AND trigger_data::text != 'null'")
    for flow_id, created, d in res.fetchall():
        if isinstance(d, dict):
            d = [d]
        rc = {'created': created, 'updated': created, 'data': d}
        ret = conn.execute(insert(rc_tbl).values(rc).returning(rc_tbl.c.id))
        d_id = ret.fetchone()[0]
        print('flow: %d  repo_changes %d' % (flow_id, d_id))
        conn.execute("UPDATE flows SET trigger_data_id = %d WHERE id = %d" % (d_id, flow_id))

    print('updating runs')
    res = conn.execute("SELECT id, created, repo_data FROM runs WHERE repo_data IS NOT NULL AND repo_data::text != 'null'")
    for run_id, created, d in res.fetchall():
        if isinstance(d, dict):
            new_ds = []
            for url, commits in d.items():
                new_d = {}
                new_d['repo'] = url
                new_d['trigger'] = 'git-push'

                new_commits = []
                for c in commits:
                    nc = dict(id=c['commit'],
                              author=dict(name=c['author'], email=c['email']),
                              timestamp=c['date'],
                              message=c['subject'])
                    new_commits.append(nc)

                new_d['commits'] = new_commits
                new_d['before'] = new_commits[-1]['id']
                new_d['after'] = new_commits[0]['id']

                new_ds.append(new_d)

            d = new_ds

        rc = {'created': created, 'updated': created, 'data': d}
        ret = conn.execute(insert(rc_tbl).values(rc).returning(rc_tbl.c.id))
        d_id = ret.fetchone()[0]
        print('run: %d  repo_changes %d' % (run_id, d_id))
        conn.execute("UPDATE runs SET repo_data_id = %d WHERE id = %d" % (d_id, run_id))

    #op.drop_column('flows', 'trigger_data')
    #op.drop_column('runs', 'repo_data')


def downgrade():
    # op.add_column('runs', sa.Column('repo_data', postgresql.JSONB(astext_type=sa.Text()), autoincrement=False, nullable=True))
    op.drop_constraint('runs_repo_data_id_fkey', 'runs', type_='foreignkey')
    op.drop_column('runs', 'repo_data_id')

    # op.add_column('flows', sa.Column('trigger_data', postgresql.JSONB(astext_type=sa.Text()), autoincrement=False, nullable=True))
    op.drop_constraint('flows_trigger_data_id_fkey', 'flows', type_='foreignkey')
    op.drop_column('flows', 'trigger_data_id')

    op.drop_table('repo_changes')



"""added casbin_rules table

Revision ID: 87358b464400
Revises: 411094f6b709
Create Date: 2022-09-30 22:52:45.199722

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '87358b464400'
down_revision = '411094f6b709'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table('casbin_rules',
                    sa.Column('created', sa.DateTime(timezone=True), nullable=False),
                    sa.Column('updated', sa.DateTime(timezone=True), nullable=False),
                    sa.Column('deleted', sa.DateTime(timezone=True), nullable=True),
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('ptype', sa.String(length=255), nullable=True),
                    sa.Column('v0', sa.String(length=255), nullable=True),
                    sa.Column('v1', sa.String(length=255), nullable=True),
                    sa.Column('v2', sa.String(length=255), nullable=True),
                    sa.Column('v3', sa.String(length=255), nullable=True),
                    sa.Column('v4', sa.String(length=255), nullable=True),
                    sa.Column('v5', sa.String(length=255), nullable=True),
                    sa.PrimaryKeyConstraint('id'))


def downgrade():
    op.drop_table('casbin_rules')



# Copyright 2020-2022 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import json
import logging
import datetime
import xmlrpc.client

from flask import abort
from sqlalchemy.orm.attributes import flag_modified
from sqlalchemy.sql.expression import asc, desc, extract
from sqlalchemy.orm import joinedload, aliased
from sqlalchemy.sql import func, select
import pytimeparse
import redis
import boto3
from azure.mgmt.subscription import SubscriptionClient
from azure.mgmt.compute import ComputeManagementClient

from . import consts, srvcheck, kkrq
from .models import db, Branch, Stage, Agent, AgentsGroup, Secret, AgentAssignment, Setting
from .models import Project, BranchSequence, System, Flow, duration_to_txt
from .models import Tool
from .schema import check_and_correct_stage_schema, SchemaError, execute_schema_code
from .schema import prepare_new_planner_triggers
from .cloud import aws, azure, k8s
from ..version import version as server_version
from . import toolutils
from . import notify
from . import utils
from . import minioops
from . import schemaval
from . import access
from . import authn
from . import chops


log = logging.getLogger(__name__)


def get_server_version():
    return {'version': server_version}, 200


def create_project(body, token_info=None):
    access.check(token_info, '', 'admin', 'only superadmin role can create projects')

    if 'name' not in body:
        abort(400, "Missing name in project")
    project = Project.query.filter_by(name=body['name']).one_or_none()
    if project is not None:
        abort(400, "Project with name %s already exists" % body['name'])

    new_project = Project(name=body['name'], description=body.get('description', ''))
    db.session.commit()

    return new_project.get_json(), 201


def update_project(project_id, body, token_info=None):
    access.check(token_info, project_id, 'pwrusr',
                 'only superadmin, project admin and project power user roles can modify project')

    project = Project.query.filter_by(id=project_id).one_or_none()
    if project is None:
        abort(404, "Project not found")

    if 'webhooks' in body:
        project.webhooks = body['webhooks']

    db.session.commit()

    result = project.get_json()
    return result, 200


def get_project(project_id, with_results, token_info=None):
    project = Project.query.filter_by(id=project_id).one_or_none()
    if project is None:
        abort(400, "Project with id %s does not exist" % project_id)

    return project.get_json(with_results=with_results), 200


def delete_project(project_id, token_info=None):
    access.check(token_info, project_id, 'admin',
                 'only superadmin and project admin roles can delete project')

    project = Project.query.filter_by(id=project_id).one_or_none()
    if project is None:
        abort(400, "Project with id %s does not exist" % project_id)

    project.deleted = utils.utcnow()
    db.session.commit()

    return {}, 200


def get_projects(token_info):
    roles = access.get_user_roles(token_info['sub'])

    q = Project.query
    q = q.filter_by(deleted=None)
    q = q.options(joinedload('branches'),
                  joinedload('branches.project'),
                  joinedload('branches.ci_last_incomplete_flow'),
                  joinedload('branches.ci_last_incomplete_flow.runs'),
                  joinedload('branches.ci_last_completed_flow'),
                  joinedload('branches.ci_last_completed_flow.runs'),
                  joinedload('secrets'))
    q = q.order_by(Project.name)
    projects = []
    for p in q.all():
        if not roles['superadmin'] and p.id not in roles['projects']:
            continue
        projects.append(p.get_json(with_last_results=True))
    return {'items': projects, 'total': len(projects)}, 200


def create_branch(project_id, body, token_info=None):
    access.check(token_info, project_id, 'pwrusr',
                 'only superadmin, project admin and project power user roles can create a branch')

    project = Project.query.filter_by(id=project_id).one_or_none()
    if project is None:
        abort(404, "Project not found")

    if 'id' in body:
        parent_branch = Branch.query.filter_by(id=body['id']).one_or_none()
        if parent_branch is None:
            abort(404, "Branch not found")
    else:
        parent_branch = None


    if 'branch_name' in body and body['branch_name']:
        branch_name = body['branch_name']
    else:
        branch_name = body['name']

    branch = Branch(project=project, name=body['name'], branch_name=branch_name)

    if parent_branch:
        if body['forking_model'] == 'model-1':
            # forked branch continues numbering, old branch resets numbering
            for bs in parent_branch.sequences:
                if bs.stage is not None:
                    continue
                BranchSequence(branch=branch, kind=bs.kind, value=bs.value)
                bs.value = 0
        else:
            # forked branch resets numbering, old branch continues numbering
            BranchSequence(branch=branch, kind=consts.BRANCH_SEQ_FLOW, value=0)
            BranchSequence(branch=branch, kind=consts.BRANCH_SEQ_CI_FLOW, value=0)
            BranchSequence(branch=branch, kind=consts.BRANCH_SEQ_DEV_FLOW, value=0)

        db.session.commit()

        # clone stages
        for stage in parent_branch.stages:
            if stage.deleted:
                continue
            new_stage = Stage(
                name=stage.name,
                description=stage.description,
                branch=branch,
                enabled=stage.enabled,
                schema=stage.schema,
                schema_code=stage.schema_code,
                timeouts=stage.timeouts,
                repo_access_token=stage.repo_access_token,
                repo_branch=stage.repo_branch,
                repo_url=stage.repo_url,
                schema_file=stage.schema_file,
                schema_from_repo_enabled=stage.schema_from_repo_enabled,
                repo_refresh_interval=stage.repo_refresh_interval,
                git_clone_params=stage.git_clone_params)

            if body['forking_model'] == 'model-1':
                # forked branch continues numbering, old branch resets numbering
                for bs in stage.sequences:
                    BranchSequence(branch=branch, stage=new_stage, kind=bs.kind, value=bs.value)
                    bs.value = 0
            else:
                # forked branch resets numbering, old branch continues numbering
                BranchSequence(branch=branch, stage=new_stage, kind=consts.BRANCH_SEQ_RUN, value=0)
                BranchSequence(branch=branch, stage=new_stage, kind=consts.BRANCH_SEQ_CI_RUN, value=0)
                BranchSequence(branch=branch, stage=new_stage, kind=consts.BRANCH_SEQ_DEV_RUN, value=0)

            db.session.flush()

            triggers = {}
            prepare_new_planner_triggers(new_stage.id, new_stage.schema['triggers'], None, triggers)
            stage.triggers = triggers

            db.session.commit()
    else:
        BranchSequence(branch=branch, kind=consts.BRANCH_SEQ_FLOW, value=0)
        BranchSequence(branch=branch, kind=consts.BRANCH_SEQ_CI_FLOW, value=0)
        BranchSequence(branch=branch, kind=consts.BRANCH_SEQ_DEV_FLOW, value=0)
        db.session.commit()

    return branch.get_json(), 201


def update_branch(branch_id, body, token_info=None):
    branch = Branch.query.filter_by(id=branch_id).one_or_none()
    if branch is None:
        abort(404, "Branch not found")

    access.check(token_info, branch.project_id, 'pwrusr',
                 'only superadmin, project admin and project power user roles can modify a branch')

    if 'name' in body:
        branch.name = body['name']

    if 'branch_name' in body:
        branch.branch_name = body['branch_name']

    db.session.commit()

    result = branch.get_json()
    return result, 200


def get_branch(branch_id, token_info=None):
    branch = Branch.query.filter_by(id=branch_id).one_or_none()
    if branch is None:
        abort(404, "Branch not found")

    access.check(token_info, branch.project_id, 'view',
                 'only superadmin, project admin, project power user and project viewer roles can get a branch')

    return branch.get_json(with_cfg=True), 200


def delete_branch(branch_id, token_info=None):
    branch = Branch.query.filter_by(id=branch_id).one_or_none()
    if branch is None:
        abort(400, "Branch with id %s does not exist" % branch_id)

    access.check(token_info, branch.project_id, 'admin',
                 'only superadmin and project admin roles can delete a branch')

    branch.deleted = utils.utcnow()
    db.session.commit()

    return {}, 200


def get_branch_sequences(branch_id, token_info=None):
    branch = Branch.query.filter_by(id=branch_id).one_or_none()
    if branch is None:
        abort(404, "Branch not found")

    access.check(token_info, branch.project_id, 'view',
                 'only superadmin, project admin, project power user and project viewer roles can get branch sequences')

    q = BranchSequence.query.filter_by(branch=branch)
    q = q.order_by(BranchSequence.id)

    seqs = []
    for bs in q.all():
        seqs.append(bs.get_json())

    return {'items': seqs, 'total': len(seqs)}, 200


def update_branch_sequence(seq_id, body, token_info=None):
    seq = BranchSequence.query.filter_by(id=seq_id).one_or_none()
    if seq is None:
        abort(404, "Branch sequence not found")

    access.check(token_info, seq.branch.project_id, 'pwrusr',
                 'only superadmin, project admin and project power user roles can get branch sequences')

    value = body.get('value', None)
    if value is None:
        abort(400, "Missing value")

    try:
        value = int(value)
    except Exception:
        abort(400, "Incorrect value")

    if value < -1:
        abort(400, "Incorrect negative value")

    seq.value = value
    db.session.commit()

    return seq.get_json(), 200


def move_branch(branch_id, body, token_info=None):
    branch = Branch.query.filter_by(id=branch_id).one_or_none()
    if branch is None:
        abort(404, "Branch not found")

    access.check(token_info, branch.project_id, 'admin',
                 'only superadmin and project admin roles can move a branch')

    proj_id = body.get('project_id', None)
    if proj_id is None:
        abort(400, "Missing project id")

    access.check(token_info, proj_id, 'admin',
                 'only superadmin and project admin roles can move a branch')

    proj = Project.query.filter_by(id=proj_id).one_or_none()
    if proj is None:
        abort(400, "Project with id %s does not exist" % proj_id)

    branch.project_id = proj.id
    db.session.commit()

    return branch.get_json(with_cfg=True), 200


def get_branch_stats(branch_id, token_info=None):
    branch = Branch.query.filter_by(id=branch_id).one_or_none()
    if branch is None:
        abort(404, "Branch not found")

    access.check(token_info, branch.project_id, 'view',
                 'only superadmin, project admin, project power user and project viewer roles can get branch sequences')

    resp = {
        'id': branch.id,
        'ci': {},
        'dev': {},
    }

    today = utils.utcnow()
    week_ago = today - datetime.timedelta(days=7)
    month_ago = today - datetime.timedelta(days=30)

    for kind, rsp in [(consts.FLOW_KIND_CI, resp['ci']), (consts.FLOW_KIND_DEV, resp['dev'])]:
        q = Flow.query.filter_by(branch=branch, kind=kind)

        # total stats
        rsp['flows_total'] = q.count()
        rsp['flows_last_month'] = q.filter(Flow.created >= month_ago).count()
        rsp['flows_last_week'] = q.filter(Flow.created >= week_ago).count()

        # duration stats
        rsp['avg_duration_last_month'] = None
        rsp['avg_duration_last_week'] = None
        rsp['durations'] = []
        if rsp['flows_last_month'] > 0:
            q2 = q.filter(Flow.finished.is_not(None))
            q2 = q2.with_entities(func.avg(extract('epoch', Flow.finished) - extract('epoch', Flow.created)).label('average'))
            secs = q2.filter(Flow.created >= month_ago).all()[0][0]
            dur = datetime.timedelta(seconds=secs)
            rsp['avg_duration_last_month'] = duration_to_txt(dur)
            if rsp['flows_last_week'] > 0:
                secs = q2.filter(Flow.created >= week_ago).all()[0][0]
                if secs:
                    dur = datetime.timedelta(seconds=secs)
                    rsp['avg_duration_last_week'] = duration_to_txt(dur)

            # durations table
            q2 = q.with_entities(Flow.id, Flow.label, extract('epoch', Flow.finished) - extract('epoch', Flow.created))
            q2 = q2.order_by(asc(Flow.created))
            durs = []
            for f_id, f_label, f_dur in q2.all():
                durs.append(dict(flow_label=f_label if f_label else ("%d." % f_id),
                                 duration=None if f_dur is None else int(f_dur)))
            rsp['durations'] = durs

    return resp, 200


def get_workflow_schema():
    schema, tools_schemas = schemaval.get_schema()
    schema["properties"]["jobs"]["items"]["properties"]["steps"]["items"]["oneOf"] = list(tools_schemas.values())
    return schema, 200


def create_secret(project_id, body, token_info=None):
    access.check(token_info, project_id, 'admin',
                 'only superadmin and project admin roles can create a secret')

    project = Project.query.filter_by(id=project_id).one_or_none()
    if project is None:
        abort(400, "Project with id %s does not exist" % project_id)

    name = body['name']
    m = re.search(r'[A-Za-z0-9_]+', name)
    if not m:
        abort(400, 'Secret name can have only A-Z, a-z, 0-9 or _ characters')

    if body['kind'] == 'ssh-key':
        kind = consts.SECRET_KIND_SSH_KEY
        data = dict(username=body['username'],
                    key=body['key'])
    elif body['kind'] == 'simple':
        kind = consts.SECRET_KIND_SIMPLE
        data = dict(secret=body['secret'])
    else:
        abort(400, "Wrong data")

    secret = Secret(project=project, name=name, kind=kind, data=data)
    db.session.commit()

    return secret.get_json(), 201


def update_secret(secret_id, body, token_info=None):
    secret = Secret.query.filter_by(id=secret_id, deleted=None).one_or_none()
    if secret is None:
        abort(404, "Secret not found")

    access.check(token_info, secret.project_id, 'admin',
                 'only superadmin and project admin roles can modify a secret')

    if 'name' in body:
        name = body['name']
        m = re.search(r'^[A-Za-z0-9_]+$', name)
        if not m:
            abort(400, 'Secret name can have only A-Z, a-z, 0-9 or _ characters')

        old_name = secret.name
        secret.name = name
        log.info('changed name from %s to %s', old_name, secret.name)

    if secret.kind == consts.SECRET_KIND_SIMPLE:
        if 'secret' in body and body['secret'] != '******':
            secret.data['secret'] = body['secret']
    elif secret.kind == consts.SECRET_KIND_SSH_KEY:
        if 'username' in body:
            secret.data['username'] = body['username']
        if 'key' in body and body['key'] != '******':
            secret.data['key'] = body['key']
    flag_modified(secret, 'data')

    db.session.commit()

    result = secret.get_json()
    return result, 200


def delete_secret(secret_id, token_info=None):
    secret = Secret.query.filter_by(id=secret_id, deleted=None).one_or_none()
    if secret is None:
        abort(404, "Secret not found")

    access.check(token_info, secret.project_id, 'admin',
                 'only superadmin and project admin roles can delete a secret')

    secret.deleted = utils.utcnow()
    db.session.commit()

    return {}, 200


def create_stage(branch_id, body, token_info=None):
    branch = Branch.query.filter_by(id=branch_id).one_or_none()
    if branch is None:
        abort(404, "Branch not found")

    access.check(token_info, branch.project_id, 'pwrusr',
                 'only superadmin, project admin and project power user roles can create a stage')

    schema_code = None
    if 'schema_code' in body:
        schema_code = body['schema_code']

    try:
        schema_code, schema = check_and_correct_stage_schema(branch, body['name'], schema_code)
    except SchemaError as e:
        abort(400, str(e))

    # create record
    stage = Stage(branch=branch, name=body['name'], schema=schema, schema_code=schema_code)
    BranchSequence(branch=branch, stage=stage, kind=consts.BRANCH_SEQ_RUN, value=0)
    BranchSequence(branch=branch, stage=stage, kind=consts.BRANCH_SEQ_CI_RUN, value=0)
    BranchSequence(branch=branch, stage=stage, kind=consts.BRANCH_SEQ_DEV_RUN, value=0)
    db.session.flush()

    triggers = {}
    prepare_new_planner_triggers(stage.id, schema['triggers'], None, triggers)
    stage.triggers = triggers

    db.session.commit()

    return stage.get_json(), 201


def get_stage(stage_id, token_info=None):
    stage = Stage.query.filter_by(id=stage_id).one_or_none()
    if stage is None:
        abort(404, "Stage not found")

    access.check(token_info, stage.branch.project_id, 'view',
                 'only superadmin, project admin, project power user and project viewer roles can get a stage')

    result = stage.get_json()
    return result, 200


def update_stage(stage_id, body, token_info=None):
    stage = Stage.query.filter_by(id=stage_id).one_or_none()
    if stage is None:
        abort(404, "Stage not found")

    access.check(token_info, stage.branch.project_id, 'pwrusr',
                 'only superadmin, project admin and project power user roles can modify a stage')

    if 'name' in body:
        stage.name = body['name']

    if 'description' in body:
        stage.description = body['description']

    if 'enabled' in body:
        stage.enabled = body['enabled']

    if 'schema_code' in body:
        prev_triggers = stage.schema['triggers']
        try:
            schema_code, schema = check_and_correct_stage_schema(stage.branch, body['name'], body['schema_code'])
        except SchemaError as e:
            abort(400, str(e))
        stage.schema = schema
        stage.schema_code = schema_code
        flag_modified(stage, 'schema')
        if stage.triggers is None:
            stage.triggers = {}
        prepare_new_planner_triggers(stage.id, schema['triggers'], prev_triggers, stage.triggers)
        flag_modified(stage, 'triggers')
        log.info('new schema: %s', stage.schema)

    if 'schema_from_repo_enabled' in body:
        schema_from_repo_enabled = body['schema_from_repo_enabled']
    else:
        schema_from_repo_enabled = stage.schema_from_repo_enabled

    stage.schema_from_repo_enabled = schema_from_repo_enabled
    db.session.commit()

    if schema_from_repo_enabled:
        if 'repo_url' in body:
            stage.repo_url = body['repo_url']
        if 'repo_branch' in body:
            stage.repo_branch = body['repo_branch']
        if 'repo_access_token' in body:
            repo_access_token = body['repo_access_token']
            secret = Secret.query.filter_by(project=stage.branch.project, name=repo_access_token).one_or_none()
            if secret is None:
                abort(400, "Secret '%s' for access token does not exist" % repo_access_token)
            if secret.kind != consts.SECRET_KIND_SIMPLE:
                abort(400, "Type of '%s' access token secret should be Simple Secret" % repo_access_token)
            stage.repo_access_token = repo_access_token

        if 'schema_file' in body:
            stage.schema_file = body['schema_file']
        if 'repo_refresh_interval' in body:
            # check if interval can be parsed
            try:
                int(body['repo_refresh_interval'])
            except Exception:
                try:
                    int(pytimeparse.parse(body['repo_refresh_interval']))
                except Exception:
                    abort(400, 'Incorrect repo refresh interval value')
            stage.repo_refresh_interval = body['repo_refresh_interval']
            log.info('stage.repo_refresh_interval %s', stage.repo_refresh_interval)
        if 'git_clone_params' in body:
            stage.git_clone_params = body['git_clone_params']

        stage.repo_state = consts.REPO_STATE_REFRESHING
        stage.repo_error = ''
        db.session.commit()

        if stage.repo_refresh_job_id:
            planner_url = os.environ.get('KRAKEN_PLANNER_URL', consts.DEFAULT_PLANNER_URL)
            planner = xmlrpc.client.ServerProxy(planner_url, allow_none=True)
            planner.remove_job(stage.repo_refresh_job_id)
            stage.repo_refresh_job_id = ''
            db.session.commit()

        from .bg import jobs as bg_jobs  # pylint: disable=import-outside-toplevel
        kkrq.enq_neck(bg_jobs.refresh_schema_repo, stage.id, 0, ignore_args=[1])

    result = stage.get_json()

    return result, 200


def delete_stage(stage_id, token_info=None):
    stage = Stage.query.filter_by(id=stage_id).one_or_none()
    if stage is None:
        abort(404, "Stage not found")

    access.check(token_info, stage.branch.project_id, 'pwrusr',
                 'only superadmin, project admin and project power user roles can delete a stage')

    stage.deleted = utils.utcnow()
    db.session.commit()

    return {}, 200


def get_stage_schema_as_json(stage_id, body, token_info=None):
    schema_code = body
    stage = Stage.query.filter_by(id=stage_id).one_or_none()
    if stage is None:
        abort(404, "Stage not found")

    access.check(token_info, stage.branch.project_id, 'view',
                 'only superadmin, project admin, project power user and project viewer roles can get a stage')

    try:
        schema = execute_schema_code(stage.branch, schema_code['schema_code'])
    except Exception as e:
        return dict(stage_id=stage_id, error=str(e)), 200

    schema = json.dumps(schema, indent=4, separators=(',', ': '))

    return dict(stage_id=stage_id, schema=schema), 200


def get_stage_schedule(stage_id, token_info=None):
    stage = Stage.query.filter_by(id=stage_id).one_or_none()
    if stage is None:
        abort(404, "Stage not found")

    access.check(token_info, stage.branch.project_id, 'view',
                 'only superadmin, project admin, project power user and project viewer roles can get a stage')

    schedules = []

    if stage.triggers:
        planner_url = os.environ.get('KRAKEN_PLANNER_URL', consts.DEFAULT_PLANNER_URL)
        planner = xmlrpc.client.ServerProxy(planner_url, allow_none=True)
        jobs = planner.get_jobs()
        jobs = {j['id']: j for j in jobs}

        for name, val in stage.triggers.items():
            if 'planner_job' not in name:
                continue
            if val not in jobs:
                continue
            job = jobs[val]
            s = dict(name=name,
                     job_id=val,
                     next_run_time=job['next_run_time'])
            schedules.append(s)

    return dict(schedules=schedules), 200


def get_agent(agent_id, token_info=None):
    access.check(token_info, '', 'admin',
                 'only superadmin can get an agent')

    ag = Agent.query.filter_by(id=agent_id).one_or_none()
    if ag is None:
        abort(400, "Cannot find agent with id %s" % agent_id)

    return ag.get_json(), 200


def get_agents(unauthorized=None, start=0, limit=10, sort_field="name", sort_dir="asc", token_info=None):
    access.check(token_info, '', 'admin',
                 'only superadmin can get agents')

    q = Agent.query
    q = q.filter_by(deleted=None)
    if unauthorized:
        q = q.filter_by(authorized=False)
    else:
        q = q.filter_by(authorized=True)

    total = q.count()

    # sorting
    sort_func = asc
    if sort_dir == "desc":
        sort_func = desc

    if sort_field in ['last_seen', 'ip_address', 'name', 'address', 'id']:
        q = q.order_by(sort_func(sort_field))
    else:
        q = q.order_by(Agent.name)

    q = q.offset(start).limit(limit)
    agents = []
    for e in q.all():
        agents.append(e.get_json())
    return {'items': agents, 'total': total}, 200


def update_agents(body, token_info=None):
    access.check(token_info, '', 'admin',
                 'only superadmin can modify agents')

    agents = body
    log.info('agents %s', agents)

    agents2 = []
    for a in agents:
        agent = Agent.query.filter_by(id=a['id']).one_or_none()
        if agent is None:
            abort(400, 'Cannot find agent %s' % a['id'])
        agents2.append(agent)

    all_group = AgentsGroup.query.filter_by(name='all').one_or_none()

    for data, agent in zip(agents, agents2):
        if 'authorized' in data:
            agent.authorized = data['authorized']

            if agent.authorized and all_group:
                already_in = False
                for ag in agent.agents_groups:
                    if ag.agents_group_id == all_group.id:
                        already_in = True
                        break
                if not already_in:
                    AgentAssignment(agent=agent, agents_group=all_group)

    db.session.commit()

    return {}, 200


def update_agent(agent_id, body, token_info=None):
    access.check(token_info, '', 'admin',
                 'only superadmin can modify an agent')

    agent = Agent.query.filter_by(id=agent_id).one_or_none()
    if agent is None:
        abort(404, "Agent not found")

    if 'groups' in body:
        # check new groups
        new_groups = set()
        if body['groups']:
            for g_id in body['groups']:
                g = AgentsGroup.query.filter_by(id=g_id['id']).one_or_none()
                if g is None:
                    abort(404, "Agents Group with id %s not found" % g_id)
                new_groups.add(g)

        # get old groups
        current_groups = set()
        assignments_map = {}
        for aa in agent.agents_groups:
            current_groups.add(aa.agents_group)
            assignments_map[aa.agents_group.id] = aa

        # remove groups
        removed = current_groups - new_groups
        for r in removed:
            aa = assignments_map[r.id]
            db.session.delete(aa)

        # add groups
        added = new_groups - current_groups
        for a in added:
            AgentAssignment(agent=agent, agents_group=a)

    if 'disabled' in body:
        agent.disabled = body['disabled']

    db.session.commit()

    return agent.get_json(), 200


def delete_agent(agent_id, token_info=None):
    access.check(token_info, '', 'admin',
                 'only superadmin can delete an agent')

    agent = Agent.query.filter_by(id=agent_id).one_or_none()
    if agent is None:
        abort(404, "Agent not found")

    if agent.job is not None:
        job = agent.job
        job.agent = None
        job.state = consts.JOB_STATE_QUEUED
        agent.job = None

    # only mark as deleted, do not unassign from groups yet as it is needed
    # in destroy_machine if it is cloud machine
    agent.deleted = utils.utcnow()
    agent.authorized = False
    agent.disabled = True
    db.session.commit()

    from .bg import jobs as bg_jobs  # pylint: disable=import-outside-toplevel
    kkrq.enq(bg_jobs.destroy_machine, agent.id)

    return {}, 200


def get_group(group_id, token_info=None):
    access.check(token_info, '', 'admin',
                 'only superadmin can get an agents group')

    ag = AgentsGroup.query.filter_by(id=group_id).one_or_none()
    if ag is None:
        abort(400, "Cannot find agent group with id %s" % group_id)

    return ag.get_json(), 200


def get_groups(start=0, limit=10, token_info=None):
    access.check(token_info, '', 'admin',
                 'only superadmin can get agents groups')

    q = AgentsGroup.query
    q = q.filter_by(deleted=None)
    q = q.order_by(AgentsGroup.name)
    total = q.count()
    q = q.offset(start).limit(limit)
    groups = []
    for ag in q.all():
        groups.append(ag.get_json())
    return {'items': groups, 'total': total}, 200


def create_group(body, token_info=None):
    access.check(token_info, '', 'admin',
                 'only superadmin can create an agents group')

    group = AgentsGroup.query.filter_by(name=body['name']).one_or_none()
    if group is not None:
        abort(400, "Group with name %s already exists" % body['name'])

    project = None
    if 'project_id' in body:
        project = Project.query.filter_by(id=body['project_id']).one_or_none()
        if project is None:
            abort(400, "Cannot find project with id %s" % body['project_id'])

    group = AgentsGroup(name=body['name'], project=project)
    db.session.commit()

    return group.get_json(), 201


def update_group(group_id, body, token_info=None):
    access.check(token_info, '', 'admin',
                 'only superadmin can modify an agents group')

    group = AgentsGroup.query.filter_by(id=group_id).one_or_none()
    if group_id is None:
        abort(404, "Group not found")

    #if 'groups' in body:
    log.info('GROUP %s', body)

    if 'name' in body:
        group.name = body['name']

    if 'deployment' in body:
        # TODO: destroy resources connected with previous deployment
        group.deployment = body['deployment']

    db.session.commit()

    return group.get_json(), 200


def delete_group(group_id, token_info=None):
    access.check(token_info, '', 'admin',
                 'only superadmin can delete an agents group')

    group = AgentsGroup.query.filter_by(id=group_id).one_or_none()
    if group is None:
        abort(404, "Agents group with id %s not found" % group_id)

    group.deleted = utils.utcnow()
    db.session.commit()

    return {}, 200


def get_aws_ec2_regions(token_info=None):
    access.check(token_info, '', 'admin',
                 'only superadmin can get aws ec2 regions')

    credential = aws.login_to_aws()
    if not credential:
        abort(500, "Incorrect AWS credential, set them in global cloud settings")
    ec2 = boto3.client('ec2', region_name='us-east-1', **credential)
    resp = ec2.describe_regions()
    return {'items': resp['Regions'], 'total': len(resp['Regions'])}, 200


def get_aws_ec2_instance_types(region, token_info=None):
    access.check(token_info, '', 'admin',
                 'only superadmin can get aws ec2 instance types')

    credential = aws.login_to_aws()
    if not credential:
        abort(500, "Incorrect AWS credential, set them in global cloud settings")
    ec2 = boto3.client('ec2', region_name=region, **credential)
    resp = ec2.describe_instance_type_offerings(Filters=[{'Name': 'location', 'Values':[region]}])
    types = resp['InstanceTypeOfferings']
    types.sort(key=lambda x: x['InstanceType'])
    return {'items': types, 'total': len(types)}, 200


def get_azure_locations(token_info=None):
    access.check(token_info, '', 'admin',
                 'only superadmin can get azure locations')

    # Acquire a credential object using service principal authentication.
    credential, subscription_id = azure.login_to_azure()
    if not credential:
        abort(500, "Incorrect Azure credential, set them in global cloud settings")
    subscription_client = SubscriptionClient(credential)
    locations = subscription_client.subscriptions.list_locations(subscription_id)
    locs = []
    for location in locations:
        locs.append(location.name)
    locs.sort()

    return {'items': locs, 'total': len(locs)}, 200


def get_azure_vm_sizes(location, token_info=None):
    access.check(token_info, '', 'admin',
                 'only superadmin can get azure vm sizes')

    # Acquire a credential object using service principal authentication.
    credential, subscription_id = azure.login_to_azure()
    if not credential:
        abort(500, "Incorrect Azure credential, set them in global cloud settings")
    compute_client = ComputeManagementClient(credential, subscription_id)
    vm_sizes_list = compute_client.virtual_machine_sizes.list(location=location)
    vm_sizes = []
    for s in vm_sizes_list:
        vm_sizes.append(s.as_dict())
    vm_sizes.sort(key=lambda x: x['name'])

    return {'items': vm_sizes, 'total': len(vm_sizes)}, 200


def _get_settings(scope=None):
    settings = Setting.query.filter_by().all()

    groups = {}
    for s in settings:
        if scope == 'public':
            if s.group not in ['idp']:
                continue
            if 'enabled' not in s.name:
                continue

        if s.group not in groups:
            groups[s.group] = {}
        grp = groups[s.group]
        grp[s.name] = s.get_value()

    return groups


def get_settings(token_info=None):
    scope = 'public'
    try:
        access.check(token_info, '', 'admin',
                     'only superadmin can get settings')
        scope = None
    except Exception:
        pass

    groups = _get_settings(scope)

    return groups, 200


def update_settings(body, token_info=None):
    access.check(token_info, '', 'admin',
                 'only superadmin can get modify settings')

    settings = Setting.query.filter_by().all()

    for group_name, group in body.items():
        for name, val in group.items():
            for s in settings:
                if s.group == group_name and s.name == name:
                    if s.val_type == 'password' and val == '':
                        continue
                    s.set_value(val)

                    if name == 'clickhouse_log_ttl':
                        ch = chops.get_clickhouse()
                        # query = 'ALTER TABLE logs DELETE WHERE time < (now() - toIntervalMonth(%(months)s))'
                        query = 'ALTER TABLE logs MODIFY TTL toDateTime(time) + toIntervalMonth(%(months)s)'
                        params = {'months': int(val)}
                        resp = ch.execute(query, params)
                        log.info('set clickhouse_log_ttl to %s, resp: %s', val, resp)

    db.session.commit()

    settings = _get_settings()
    return settings, 200


def get_diagnostics(token_info=None):
    access.check(token_info, '', 'admin',
                 'only superadmin can get diagnostics')

    diags = {}

    # check postgresql
    pgsql_addr = os.environ.get('KRAKEN_DB_URL', consts.DEFAULT_DB_URL)
    pgsql_open = srvcheck.is_service_open(pgsql_addr)
    diags['postgresql'] = {
        'name': 'PostgreSQL',
        'address': pgsql_addr,
        'open': pgsql_open
    }

    # check clickhouse
    ch_url = chops.get_clickhouse_url()
    ch_open = srvcheck.is_service_open(ch_url)
    diags['clickhouse'] = {
        'name': 'ClickHouse',
        'address': ch_url,
        'open': ch_open
    }

    # check redis
    rds_addr = os.environ.get('KRAKEN_REDIS_ADDR', consts.DEFAULT_REDIS_ADDR)
    rds_open = srvcheck.is_service_open(rds_addr, 6379)
    diags['redis'] = {
        'name': 'Redis',
        'address': rds_addr,
        'open': rds_open
    }

    # check minio
    minio_addr, _, _ = minioops.get_minio_addr()
    minio_open = srvcheck.is_service_open(minio_addr)
    if minio_open:
        minio_open = minioops.check_connection()
    diags['minio'] = {
        'name': 'MinIO',
        'address': minio_addr,
        'open': minio_open
    }

    # check planner
    plnr_addr = os.environ.get('KRAKEN_PLANNER_URL', consts.DEFAULT_PLANNER_URL)
    plnr_open = srvcheck.is_service_open(plnr_addr)
    diags['planner'] = {
        'name': 'Kraken Planner',
        'address': plnr_addr,
        'open': plnr_open
    }

    # rq overview
    diags['rq'] = {
        'name': 'RQ',
        'address': '',
        'open': True,
    }

    # get current RQ jobs
    now = datetime.datetime.now()
    week_ago = now - datetime.timedelta(days=7)
    all_jobs = kkrq.get_jobs()
    for jobs, name in zip(all_jobs, ['scheduled', 'current', 'finished', 'failed', 'deferred']):
        jobs2 = []
        for job in jobs:
            if job.ended_at and job.ended_at < week_ago:
                continue
            job = dict(id=job.id,
                       created_at=job.created_at,
                       ended_at=job.ended_at,
                       enqueued_at=job.enqueued_at,
                       func_name=job.func_name,
                       description=job.description,
                       status=job.get_status(refresh=False))
            jobs2.append(job)
        key = '%s_jobs' % name
        diags['rq'][key] = jobs2

    return diags


def get_last_rq_jobs_names(token_info=None):
    access.check(token_info, '', 'admin',
                 'only superadmin can get last rq job names')

    # get the last RQ jobs
    ch = chops.get_clickhouse()

    now = utils.utcnow()
    start_date = now - datetime.timedelta(hours=12111)
    query = "select min(time) as mt, tool, count(*) from logs "
    query += "where service = 'rq' and tool != '' "
    query += "group by tool "
    query += "having mt > %(start_date)s "
    query += "order by mt desc "
    query += "limit 100;"
    resp = ch.execute(query, {'start_date': start_date})
    task_names = []
    for row in resp:
        task_names.append(dict(time=row[0], name=row[1], lines=row[2]))

    return {'items': task_names}, 200


def get_services_logs(services, level=None, token_info=None):
    access.check(token_info, '', 'admin',
                 'only superadmin can get services logs')

    ch = chops.get_clickhouse()

    query = "select time,message,service,host,level,tool from logs "
    where = []
    params = {}
    for idx, s in enumerate(services):
        if s == 'all':
            continue
        param = 'service%d' % idx
        if '/' in s:
            s, t = s.split('/')
            tparam = 'tool%d' % idx
            where.append("(service = %%(%s)s and tool = %%(%s)s)" % (param, tparam))
            params[param] = s
            params[tparam] = t
        else:
            where.append("service = %%(%s)s" % param)
            params[param] = s
    if where:
        where = " or ".join(where)
        where = "where (" + where + ") "
    if level:
        level = level.upper()
        if level == 'ERROR':
            lq = "level = 'ERROR'"
        elif level == 'WARNING':
            lq = "level in ('WARNING', 'ERROR')"
        else:
            lq = "level in ('INFO', 'WARNING', 'ERROR')"
        if where:
            where += "and %s " % lq
        else:
            where = "where %s " % lq
    if where:
        query += where
    query += " order by time desc, seq desc limit 1000"
    rows = ch.execute(query, params)

    logs = []
    for r in reversed(rows):
        entry = dict(time=r[0],
                     message=r[1],
                     service=r[2],
                     host=r[3],
                     level=r[4],
                     tool=r[5])
        logs.append(entry)

    return {'items': logs, 'total': len(logs)}, 200


def get_live_data():
    redis_addr = os.environ.get('KRAKEN_REDIS_ADDR', consts.DEFAULT_REDIS_ADDR)
    redis_host, redis_port = utils.split_host_port(redis_addr, 6379)
    rds = redis.Redis(host=redis_host, port=redis_port, db=consts.REDIS_KRAKEN_DB)

    counters = {
        'error_logs_count': 0,
        'authorized_agents': 0,
        'non_authorized_agents': 0,
    }

    cntr_names = list(counters.keys())
    for cntr in cntr_names:
        rds_cntr = cntr.replace('_', '-')
        cnt = rds.get(rds_cntr)
        if cnt:
            counters[cntr] = int(cnt)

    return counters, 200


def get_settings_working_state(resource, token_info=None):
    access.check(token_info, '', 'admin',
                 'only superadmin can get settings working state')

    if resource == 'email':
        state = notify.check_email_settings()
    elif resource == 'slack':
        state = notify.check_slack_settings()
    elif resource == 'aws':
        state = aws.check_aws_settings()
    elif resource == 'azure':
        state = azure.check_azure_settings()
    elif resource == 'kubernetes':
        state = k8s.check_k8s_settings()
    elif resource == 'ldap':
        state = authn.check_ldap_settings()
    else:
        abort(400, "Unsupported resource type: %s" % resource)

    return {'state': state}, 200


def get_systems():
    q = System.query
    q = q.order_by(System.name)

    systems = []
    for s in q.all():
        systems.append(s.get_json())
    return {'items': systems, 'total': len(systems)}, 200


def get_tools(start=0, limit=30, sort_field="name", sort_dir="asc"):
    q = Tool.query
    q = q.distinct(Tool.name)
    q = q.filter_by(deleted=None)
    q = q.order_by(Tool.name)

    total = q.count()

    # distinct does not allow for arbitrary sorting so wrap it in subquery
    sq = q.subquery()
    aliased_tool = aliased(Tool, sq)

    q = select(aliased_tool)

    sort_func = asc
    if sort_dir == "desc":
        sort_func = desc

    if sort_field in ['location', 'entry', 'name', 'version', 'id']:
        q = q.order_by(sort_func(sort_field))

    q = q.offset(start).limit(limit)

    tools = []
    for t in db.session.execute(q):
        t = t[0]
        tools.append(t.get_json(with_details=True))
    return {'items': tools, 'total': total}, 200


def get_tool_versions(name, start=0, limit=30, sort_field="name", sort_dir="asc"):
    q = Tool.query
    q = q.filter_by(name=name)
    q = q.filter_by(deleted=None)

    total = q.count()

    sort_func = asc
    if sort_dir == "desc":
        sort_func = desc

    if sort_field in ['location', 'entry', 'name', 'version', 'id']:
        q = q.order_by(sort_func(sort_field))
        q = q.order_by(desc(Tool.created))
    else:
        q = q.order_by(sort_func(Tool.created))

    q = q.offset(start).limit(limit)

    tools = []
    for t in q.all():
        tools.append(t.get_json(with_details=True))
    return {'items': tools, 'total': total}, 200


def _create_remote_tool(body):
    url = body['url']
    tag = body['tag']
    tool_file = body['tool_file']

    # fint the latest tool to estimate next version
    q = Tool.query
    q = q.filter_by(url=url)
    q = q.filter_by(tag=tag)
    q = q.filter_by(tool_file=tool_file)
    q = q.filter_by(deleted=None)
    tool = q.one_or_none()

    if tool:
        return tool

    tool = Tool(name=url, version=tag, url=url, tag=tag, tool_file=tool_file, fields={})
    db.session.commit()

    return tool


def create_or_update_tool(body, token_info=None):
    access.check(token_info, '', 'admin',
                 'only superadmin can create or update a tool')

    if 'name' in body:
        tool = toolutils.create_or_update_tool(body)
        db.session.commit()

    elif 'url' in body:
        tool = _create_remote_tool(body)

        from .bg import jobs as bg_jobs  # pylint: disable=import-outside-toplevel
        kkrq.enq_neck(bg_jobs.load_remote_tool, tool.id)
    else:
        abort(400, 'missing name or url in body')

    return tool.get_json(with_details=True), 201


def upload_new_or_overwrite_tool(name, body, file=None, token_info=None):
    access.check(token_info, '', 'admin',
                 'only superadmin can upload a tool')

    meta = body['meta']
    if name != meta['name']:
        msg = 'Name in description (%s) does not match name in url (%s)' % (meta['name'], name)
        abort(400, msg)

    tool = toolutils.create_or_update_tool(meta)
    db.session.flush()

    toolutils.store_tool_in_minio(file.stream, tool)
    db.session.commit()

    return tool.get_json(with_details=True), 201


def delete_tool(tool_id, token_info=None):
    access.check(token_info, '', 'admin',
                 'only superadmin can delete a tool')

    tool = Tool.query.filter_by(id=tool_id).one_or_none()

    if tool is None:
        abort(404, "Tool with id %s not found" % tool_id)

    if not tool.location and not tool.entry:
        abort(400, "Cannot archive built-in tool")

    tool.deleted = utils.utcnow()
    db.session.commit()

    return {}, 200


def update_branch_retention_policy(branch_id, body, token_info=None):
    branch = Branch.query.filter_by(id=branch_id).one_or_none()
    if branch is None:
        abort(404, "Branch not found")

    access.check(token_info, branch.project_id, 'pwrusr',
                 'only superadmin, project admin and project power user roles can update branch retention policy')

    branch.retention_policy = body
    flag_modified(branch, 'retention_policy')
    db.session.commit()

    return branch.retention_policy, 200
